# sage_setup: distribution = sagemath-ntl
# distutils: libraries = NTL_LIBRARIES gmp
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++
# sage.doctest: needs sage.libs.pari
r"""
Dense univariate polynomials over  `\ZZ/n\ZZ`, implemented using NTL

This implementation is generally slower than the FLINT implementation in
:mod:`~sage.rings.polynomial.polynomial_zmod_flint`, so we use FLINT by
default when the modulus is small enough; but NTL does not require that `n` be
``int``-sized, so we use it as default when `n` is too large for FLINT.

Note that the classes :class:`Polynomial_dense_modn_ntl_zz` and
:class:`Polynomial_dense_modn_ntl_ZZ` are different; the former is limited to
moduli less than a certain bound, while the latter supports arbitrarily large
moduli.

AUTHORS:

- Robert Bradshaw: Split off from polynomial_element_generic.py (2007-09)
- Robert Bradshaw: Major rewrite to use NTL directly (2007-09)
"""
# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on, sig_off

from sage.rings.polynomial.polynomial_element cimport Polynomial, _dict_to_list

from sage.rings.integer cimport smallInteger

from sage.libs.ntl.ntl_ZZX import ntl_ZZX as ZZX
from sage.libs.ntl.ntl_ZZ_pX import ntl_ZZ_pX as ZZ_pX
from sage.rings.integer_ring import ZZ

from sage.rings.fraction_field_element import FractionFieldElement

from sage.rings.infinity import infinity

from sage.structure.element import canonical_coercion, coerce_binop, have_same_parent

from sage.libs.ntl.types cimport NTL_SP_BOUND
from sage.libs.ntl.ZZ_p cimport *
from sage.libs.ntl.lzz_p cimport *
from sage.libs.ntl.lzz_pX cimport *
from sage.libs.ntl.ZZ_pX cimport *

try:
    from sage.libs.pari import pari
    from cypari2.gen import Gen as pari_gen
except ImportError:
    pari_gen = ()


def make_element(parent, args):
    return parent(*args)


zz_p_max = NTL_SP_BOUND

cdef class Polynomial_dense_mod_n(Polynomial):
    """
    A dense polynomial over the integers modulo n, where n is composite, with
    the underlying arithmetic done using NTL.

    EXAMPLES::

        sage: R.<x> = PolynomialRing(Integers(16), implementation='NTL')
        sage: f = x^3 - x + 17
        sage: f^2
        x^6 + 14*x^4 + 2*x^3 + x^2 + 14*x + 1

        sage: loads(f.dumps()) == f
        True

        sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
        sage: p = 3*x
        sage: q = 7*x
        sage: p + q
        10*x
        sage: R.<x> = PolynomialRing(Integers(8), implementation='NTL')
        sage: parent(p)
        Univariate Polynomial Ring in x over Ring of integers modulo 100 (using NTL)
        sage: p + q
        10*x
        sage: R({10:-1})
        7*x^10

    TESTS::

        sage: # needs sage.libs.pari
        sage: f = Integers(5*2^100)['x'].random_element()
        sage: from sage.rings.polynomial.polynomial_modn_dense_ntl import Polynomial_dense_mod_n
        sage: isinstance(f, Polynomial_dense_mod_n)
        True
    """
    def __init__(self, parent, x=None, check=True,
                 is_gen=False, construct=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
        cdef Polynomial_dense_mod_n numer, denom

        if construct:
            if isinstance(x, ZZ_pX):
                self._poly = x
                return
            self._poly = ZZ_pX(x, parent.modulus())
            return

        self._poly = ZZ_pX([], parent.modulus())

        if x is None:
            return         # leave initialized to 0 polynomial.

        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                self._poly = (<Polynomial_dense_modn_ntl_zz>x)._poly.__copy__()
                return
            else:
                R = parent.base_ring()
                x = [ZZ(R(a)) for a in x.list()]
                check = False

        elif isinstance(x, dict):
            R = parent.base_ring()
            x = _dict_to_list(x, R(0))

        elif isinstance(x, ZZX):
            self._poly = x.copy()
            return

        elif isinstance(x, pari_gen):
            x = [ZZ(w) for w in x.list()]
            check = False

        elif isinstance(x, FractionFieldElement) and \
                 isinstance(x.numerator(), Polynomial_dense_mod_n):
            if x.denominator().is_unit():
                numer = x.numerator()
                denom = x.denominator().inverse_of_unit()
                x = numer._poly * denom._poly
                check = False
            else:
                raise TypeError("Denominator not a unit.")

        elif not isinstance(x, list) and not isinstance(x, tuple):
            x = [x]   # constant polynomials

        if check:
            R = parent.base_ring()
            x = [ZZ(R(a)) for a in x]

        self._poly = ZZ_pX(x, parent.modulus())

    def __reduce__(self):
        return make_element, (self.parent(), (self.list(), False, self.is_gen()))

    def int_list(self):
        return eval(str(self._poly).replace(' ',','))

    def __pari__(self, variable=None):
        """
        EXAMPLES::

            sage: t = PolynomialRing(IntegerModRing(17), "t", implementation='NTL').gen()
            sage: f = t^3 + 3*t - 17
            sage: pari(f)                                                               # needs sage.libs.pari
            Mod(1, 17)*t^3 + Mod(3, 17)*t
        """
        if variable is None:
            variable = self.parent().variable_name()
        return pari(self.int_list()).Polrev(variable) * \
               pari(1).Mod(self.parent().base_ring().order())

    def ntl_ZZ_pX(self):
        r"""
        Return underlying NTL representation of this polynomial.  Additional
        ''bonus'' functionality is available through this function.

        .. warning::

            You must call ``ntl.set_modulus(ntl.ZZ(n))`` before doing
            arithmetic with this object!
        """
        return self._poly

    cdef get_unsafe(self, Py_ssize_t n):
        """
        Return the `n`-th coefficient of ``self``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: from sage.rings.polynomial.polynomial_modn_dense_ntl import Polynomial_dense_mod_n
            sage: f = Polynomial_dense_mod_n(R, [5,10,13,1,4]); f
            4*x^4 + x^3 + 13*x^2 + 10*x + 5
            sage: f[2]
            13
            sage: f[:3]
            13*x^2 + 10*x + 5
        """
        return self._parent._base((<ntl_ZZ_pX> self._poly)[n]._integer_())

    def _unsafe_mutate(self, n, value):
        n = int(n)
        if n < 0:
            raise IndexError("n must be >= 0")
        self._poly[n] = int(value)

    def _pow(self, n):
        n = int(n)
        if self.degree() <= 0:
            return self.parent()(self[0]**n)
        if n < 0:
            return (~self)**(-n)
        return self.parent()(self._poly**n, construct=True)

    cpdef _add_(self, right):
        return self.parent()(self._poly + (<Polynomial_dense_mod_n>right)._poly, construct=True)

    cpdef _mul_(self, right):
        """
        EXAMPLES::

            sage: x = PolynomialRing(Integers(100), 'x', implementation='NTL').0
            sage: (x - 2)*(x^2 - 8*x + 16)
            x^3 + 90*x^2 + 32*x + 68
        """
        return self.parent()(self._poly * (<Polynomial_dense_mod_n>right)._poly, construct=True)

    cpdef _lmul_(self, Element c):
        try:
            return self.parent()(ZZ_pX([c], self.parent().modulus()) * self._poly, construct=True)
        except RuntimeError as msg: # should this really be a TypeError
            raise TypeError(msg)

    @coerce_binop
    def quo_rem(self, right):
        """
        Return a tuple ``(quotient, remainder)`` where ``self = quotient*other +
        remainder``.
        """
        v = self._poly.quo_rem((<Polynomial_dense_mod_n>right)._poly)
        P = self.parent()
        return (P(v[0], construct=True), P(v[1], construct=True) )

    def shift(self, n):
        r"""
        Return this polynomial multiplied by the power `x^n`. If `n` is negative,
        terms below `x^n` will be discarded. Does not change this polynomial.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(12345678901234567890), implementation='NTL')
            sage: p = x^2 + 2*x + 4
            sage: p.shift(0)
             x^2 + 2*x + 4
            sage: p.shift(-1)
             x + 2
            sage: p.shift(-5)
             0
            sage: p.shift(2)
             x^4 + 2*x^3 + 4*x^2

        TESTS::

            sage: p = R(0)
            sage: p.shift(3).is_zero()
            True
            sage: p.shift(-3).is_zero()
            True

        AUTHOR:

        - David Harvey (2006-08-06)
        """
        if n == 0 or self.degree() < 0:
            return self
        return self.parent()(self._poly.left_shift(n),
                             construct=True)

    cpdef _sub_(self, right):
        return self.parent()(self._poly - (<Polynomial_dense_mod_n>right)._poly, construct=True)

    def __floordiv__(self, right):
        q, _ = self.quo_rem(right)
        return q

    def degree(self, gen=None):
        """
        Return the degree of this polynomial.

        The zero polynomial has degree -1.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: (x^3 + 3*x - 17).degree()
            3
            sage: R.zero().degree()
            -1

        TESTS:

        Check output type (see :issue:`25182`)::

            sage: R.<x> = PolynomialRing(Integers(3), implementation='NTL')
            sage: isinstance(x.degree(), Integer)
            True
        """
        return smallInteger(max(self._poly.degree(), -1))

    cpdef list list(self, bint copy=True):
        """
        Return a new copy of the list of the underlying
        elements of ``self``.

        EXAMPLES::

            sage: _.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: f = x^3 + 3*x - 17
            sage: f.list()
            [83, 3, 0, 1]
        """
        return [self.get_unsafe(n) for n from 0 <= n <= self.degree()]

    def ntl_set_directly(self, v):
        r"""
        Set the value of this polynomial directly from a vector or string.

        Polynomials over the integers modulo n are stored internally using
        NTL's ``ZZ_pX`` class.  Use this function to set the value of this
        polynomial using the NTL constructor, which is potentially *very* fast.
        The input v is either a vector of ints or a string of the form ``[ n1
        n2 n3 ... ]`` where the ni are integers and there are no commas between
        them. The optimal input format is the string format, since that's what
        NTL uses by default.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: from sage.rings.polynomial.polynomial_modn_dense_ntl import Polynomial_dense_mod_n as poly_modn_dense
            sage: poly_modn_dense(R, ([1,-2,3]))
            3*x^2 + 98*x + 1
            sage: f = poly_modn_dense(R, 0)
            sage: f.ntl_set_directly([1,-2,3])
            sage: f
            3*x^2 + 98*x + 1
            sage: f.ntl_set_directly('[1 -2 3 4]')
            sage: f
            4*x^3 + 3*x^2 + 98*x + 1
        """
        if self.is_gen():
            raise TypeError("Cannot change the value of the generator.")
        self._poly = ZZ_pX(v, self.parent().modulus())

    # Polynomial_singular_repr stuff, copied due to lack of multiple inheritance
    def _singular_(self, singular=None, force=False):
        self.parent()._singular_(singular, force=force).set_ring()  # this is expensive
        if self.__singular is not None:
            try:
                self.__singular._check_valid()
                if self.__singular.parent() is singular:
                    return self.__singular
            except (AttributeError, ValueError):
                pass
        return self._singular_init_(singular)

    def _singular_init_(self, singular=None, force=False):
        self.parent()._singular_(singular, force=force).set_ring()  # this is expensive
        self.__singular = singular(str(self))
        return self.__singular

    @coerce_binop
    def minpoly_mod(self, other):
        r"""
        Compute the minimal polynomial of this polynomial modulo another
        polynomial in the same ring.

        ALGORITHM:

        NTL's ``MinPolyMod()``, which uses Shoup's algorithm [Sho1999]_.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(GF(101), implementation='NTL')
            sage: f = x^17 + x^2 - 1
            sage: (x^2).minpoly_mod(f)
            x^17 + 100*x^2 + 2*x + 100

        TESTS:

        Random testing::

            sage: # needs sage.libs.pari
            sage: p = random_prime(2^99)
            sage: R.<x> = PolynomialRing(GF(p), implementation='NTL')
            sage: d = randrange(1,50)
            sage: f = R.random_element(d)
            sage: g = R.random_element((-1,5*d))
            sage: poly = g.minpoly_mod(f)
            sage: poly(R.quotient(f)(g))
            0
        """
        mod = other.ntl_ZZ_pX()
        elt = self.ntl_ZZ_pX() % mod
        res = elt.minpoly_mod(mod)
        return self.parent()(res, construct=True)

    def small_roots(self, *args, **kwds):
        r"""
        See :func:`sage.rings.polynomial.polynomial_modn_dense_ntl.small_roots`
        for the documentation of this function.

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: N = 10001
            sage: K = Zmod(10001)
            sage: P.<x> = PolynomialRing(K, implementation='NTL')
            sage: f = x^3 + 10*x^2 + 5000*x - 222
            sage: f.small_roots()                                                       # needs fpylll
            [4]
        """
        return small_roots(self, *args, **kwds)

    def compose_mod(self, other, modulus):
        r"""
        Compute `f(g) \pmod h`.

        To be precise about the order fo compostion, given ``self``, ``other``
        and ``modulus`` as `f(x)`, `g(x)` and `h(x)` compute `f(g(x)) \bmod h(x)`.

        INPUT:

        - ``other`` -- a polynomial `g(x)`
        - ``modulus`` -- a polynomial `h(x)`

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: R.<x> = GF(2**127 - 1)[]
            sage: f = R.random_element()
            sage: g = R.random_element()
            sage: g.compose_mod(g, f) == g(g) % f
            True

            sage: R.<x> = GF(163)[]
            sage: f = R([i for i in range(100)])
            sage: g = R([i**2 for i in range(100)])
            sage: h = 1 + x + x**5
            sage: f.compose_mod(g, h)
            82*x^4 + 56*x^3 + 45*x^2 + 60*x + 127
            sage: f.compose_mod(g, h) == f(g) % h
            True

        AUTHORS:

        - Giacomo Pope (2024-08) initial implementation
        """
        elt = self.ntl_ZZ_pX()
        mod = modulus.ntl_ZZ_pX()
        other = other.ntl_ZZ_pX()
        res = elt.compose_mod(other, mod)
        return self.parent()(res, construct=True)

    # compose_mod is the natural name from the NTL bindings, but polynomial_gf2x
    # has modular_composition as the method name so here we allow both
    modular_composition = compose_mod


def small_roots(self, X=None, beta=1.0, epsilon=None, **kwds):
    r"""
    Let `N` be the characteristic of the base ring this polynomial
    is defined over: ``N = self.base_ring().characteristic()``.
    This method returns small roots of this polynomial modulo some
    factor `b` of `N` with the constraint that `b >= N^\beta`.
    Small in this context means that if `x` is a root of `f`
    modulo `b` then `|x| < X`. This `X` is either provided by the
    user or the maximum `X` is chosen such that this algorithm
    terminates in polynomial time. If `X` is chosen automatically
    it is `X = ceil(1/2 N^{\beta^2/\delta - \epsilon})`.
    The algorithm may also return some roots which are larger than `X`.
    'This algorithm' in this context means Coppersmith's algorithm for finding
    small roots using the LLL algorithm. The implementation of this algorithm
    follows Alexander May's PhD thesis referenced below.

    INPUT:

    - ``X`` -- an absolute bound for the root (default: see above)
    - ``beta`` -- compute a root mod `b` where `b` is a factor of `N` and `b
      \ge N^\beta` (default: 1.0, so `b = N`.)
    - ``epsilon`` -- the parameter `\epsilon` described above. (default: `\beta/8`)
    - ``**kwds`` -- passed through to method :meth:`Matrix_integer_dense.LLL() <sage.matrix.matrix_integer_dense.Matrix_integer_dense.LLL>`

    EXAMPLES:

    First consider a small example::

        sage: # needs sage.libs.pari
        sage: N = 10001
        sage: K = Zmod(10001)
        sage: P.<x> = PolynomialRing(K, implementation='NTL')
        sage: f = x^3 + 10*x^2 + 5000*x - 222

    This polynomial has no roots without modular reduction (i.e. over `\ZZ`)::

        sage: f.change_ring(ZZ).roots()                                                 # needs sage.libs.pari
        []

    To compute its roots we need to factor the modulus `N` and use the Chinese
    remainder theorem::

        sage: # needs sage.libs.pari
        sage: p, q = N.prime_divisors()
        sage: f.change_ring(GF(p)).roots()
        [(4, 1)]
        sage: f.change_ring(GF(q)).roots()
        [(4, 1)]
        sage: crt(4, 4, p, q)
        4

    This root is quite small compared to `N`, so we can attempt to
    recover it without factoring `N` using Coppersmith's small root
    method::

        sage: f.small_roots()                                                           # needs fpylll sage.libs.pari
        [4]

    An application of this method is to consider RSA. We are using 512-bit RSA
    with public exponent `e=3` to encrypt a 56-bit DES key. Because it would be
    easy to attack this setting if no padding was used we pad the key `K` with
    1s to get a large number::

        sage: Nbits, Kbits = 512, 56
        sage: e = 3

    We choose two primes of size 256-bit each::

        sage: p = 2^256 + 2^8 + 2^5 + 2^3 + 1
        sage: q = 2^256 + 2^8 + 2^5 + 2^3 + 2^2 + 1
        sage: N = p*q
        sage: ZmodN = Zmod( N )

    We choose a random key::

        sage: K = ZZ.random_element(0, 2^Kbits)

    and pad it with `512-56=456` 1s::

        sage: Kdigits = K.digits(2)
        sage: M = [0]*Kbits + [1]*(Nbits-Kbits)
        sage: for i in range(len(Kdigits)): M[i] = Kdigits[i]

        sage: M = ZZ(M, 2)

    Now we encrypt the resulting message::

        sage: C = ZmodN(M)^e

    To recover `K` we consider the following polynomial modulo `N`::

        sage: P.<x> = PolynomialRing(ZmodN, implementation='NTL')                       # needs sage.libs.pari
        sage: f = (2^Nbits - 2^Kbits + x)^e - C                                         # needs sage.libs.pari

    and recover its small roots::

        sage: Kbar = f.small_roots()[0]                                                 # needs fpylll sage.libs.pari
        sage: K == Kbar                                                                 # needs fpylll sage.libs.pari
        True

    The same algorithm can be used to factor `N = pq` if partial
    knowledge about `q` is available. This example is from the
    Magma handbook:

    First, we set up `p`, `q` and `N`::

        sage: length = 512
        sage: hidden = 110
        sage: p = next_prime(2^int(round(length/2)))
        sage: q = next_prime(round(pi.n()*p))                                           # needs sage.symbolic
        sage: N = p*q                                                                   # needs sage.symbolic

    Now we disturb the low 110 bits of `q`::

        sage: qbar = q + ZZ.random_element(0, 2^hidden - 1)                             # needs sage.symbolic

    And try to recover `q` from it::

        sage: F.<x> = PolynomialRing(Zmod(N), implementation='NTL')                     # needs sage.symbolic
        sage: f = x - qbar                                                              # needs sage.symbolic

    We know that the error is `\le 2^{\text{hidden}}-1` and that the modulus
    we are looking for is `\ge \sqrt{N}`::

        sage: from sage.misc.verbose import set_verbose
        sage: set_verbose(2)
        sage: d = f.small_roots(X=2^hidden-1, beta=0.5)[0]  # time random               # needs fpylll sage.symbolic
        verbose 2 (<module>) m = 4
        verbose 2 (<module>) t = 4
        verbose 2 (<module>) X = 1298074214633706907132624082305023
        verbose 1 (<module>) LLL of 8x8 matrix (algorithm fpLLL:wrapper)
        verbose 1 (<module>) LLL finished (time = 0.006998)
        sage: q == qbar - d                                                             # needs fpylll sage.symbolic
        True

    REFERENCES:

    Don Coppersmith. *Finding a small root of a univariate modular equation.*
    In Advances in Cryptology, EuroCrypt 1996, volume 1070 of Lecture
    Notes in Computer Science, p. 155--165. Springer, 1996.
    http://cr.yp.to/bib/2001/coppersmith.pdf

    Alexander May. *New RSA Vulnerabilities Using Lattice Reduction Methods.*
    PhD thesis, University of Paderborn, 2003.
    http://www.cs.uni-paderborn.de/uploads/tx_sibibtex/bp.pdf
    """
    from sage.misc.verbose import verbose
    from sage.matrix.constructor import Matrix
    from sage.rings.real_mpfr import RR

    N = self.parent().characteristic()

    if not self.is_monic():
        raise ArithmeticError("Polynomial must be monic.")

    beta = RR(beta)
    if beta <= 0.0 or beta > 1.0:
        raise ValueError("0.0 < beta <= 1.0 not satisfied.")

    f = self.change_ring(ZZ)

    _, (x,) = f.parent().objgens()

    delta = f.degree()

    if epsilon is None:
        epsilon = beta/8
    verbose("epsilon = %f" % epsilon, level=2)

    m = max(beta**2/(delta * epsilon), 7*beta/delta).ceil()
    verbose("m = %d" % m, level=2)

    t = int( ( delta*m*(1/beta -1) ).floor() )
    verbose("t = %d" % t, level=2)

    if X is None:
        X = (0.5 * N**(beta**2/delta - epsilon)).ceil()
    verbose("X = %s" % X, level=2)

    # we could do this much faster, but this is a cheap step
    # compared to LLL
    g  = [x**j * N**(m-i) * f**i for i in range(m) for j in range(delta)]
    g.extend([x**i * f**m for i in range(t)])  # h

    B = Matrix(ZZ, len(g), delta*m + max(delta,t) )
    for i in range(B.nrows()):
        for j in range( g[i].degree()+1 ):
            B[i, j] = g[i][j]*X**j

    B = B.LLL(**kwds)

    f = sum([ZZ(B[0, i]//X**i)*x**i for i in range(B.ncols())])
    R = f.roots()

    ZmodN = self.base_ring()
    roots = set(ZmodN(r) for r, m in R if abs(r) <= X)
    Nbeta = N**beta
    return [root for root in roots if N.gcd(ZZ(self(root))) >= Nbeta]


cdef class Polynomial_dense_modn_ntl_zz(Polynomial_dense_mod_n):
    r"""
    Polynomial on `\ZZ/n\ZZ` implemented via NTL.

    .. automethod:: _add_
    .. automethod:: _sub_
    .. automethod:: _lmul_
    .. automethod:: _rmul_
    .. automethod:: _mul_
    .. automethod:: _mul_trunc_
    """
    def __init__(self, parent, v=None, check=True, is_gen=False, construct=False):
        r"""
        EXAMPLES::

            sage: R = Integers(5**21)
            sage: S.<x> = PolynomialRing(R, implementation='NTL')                       # needs sage.libs.pari
            sage: S(1/4)                                                                # needs sage.libs.pari
            357627868652344
        """
        if isinstance(v, Polynomial):
            if (<Element>v)._parent == parent:
                Polynomial.__init__(self, parent, is_gen=is_gen)
                self.x = (<Polynomial_dense_modn_ntl_zz>v).x
                self.c = (<Polynomial_dense_modn_ntl_zz>v).c
                return

        Polynomial_dense_mod_n.__init__(self, parent, v, check=check, is_gen=is_gen, construct=construct)
        v = [a for a in self._poly.list()]
        self._poly = None # this will eventually go away
        cdef ntl_zz_pX ntl = ntl_zz_pX(v, parent.modulus()) # let it handle the hard work
        self.x = ntl.x
        self.c = ntl.c

    def __dealloc__(self):
        if <object>self.c is not None:
            self.c.restore_c()

    def ntl_set_directly(self, v):
        # TODO: Get rid of this
        Polynomial_dense_mod_n.ntl_set_directly(self, v)
        # verbatim from __init__
        v = [int(a) for a in self._poly.list()]
        self._poly = None # this will eventually go away
        cdef ntl_zz_pX ntl = ntl_zz_pX(v, self._parent.modulus()) # let it handle the hard work
        self.x = ntl.x
        self.c = ntl.c

    cdef Polynomial_dense_modn_ntl_zz _new(self):
        cdef Polynomial_dense_modn_ntl_zz y = <Polynomial_dense_modn_ntl_zz>Polynomial_dense_modn_ntl_zz.__new__(Polynomial_dense_modn_ntl_zz)
        y.c = self.c
        y._parent = self._parent
        return y

    def int_list(self):
        """
        Return the coefficients of ``self`` as efficiently as possible as a
        list of python ints.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: from sage.rings.polynomial.polynomial_modn_dense_ntl import Polynomial_dense_mod_n as poly_modn_dense
            sage: f = poly_modn_dense(R,[5,0,0,1])
            sage: f.int_list()
            [5, 0, 0, 1]
            sage: [type(a) for a in f.int_list()]
            [<... 'int'>, <... 'int'>, <... 'int'>, <... 'int'>]
        """
        cdef long i
        return [ zz_p_rep(zz_pX_GetCoeff(self.x, i)) for i from 0 <= i <= zz_pX_deg(self.x) ]

    cdef get_unsafe(self, Py_ssize_t n):
        """
        Return the `n`-th coefficient of ``self``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: from sage.rings.polynomial.polynomial_modn_dense_ntl import Polynomial_dense_modn_ntl_zz
            sage: f = Polynomial_dense_modn_ntl_zz(R,[2, 1])^7
            sage: f[3]
            60
            sage: f[:6]
            84*x^5 + 80*x^4 + 60*x^3 + 72*x^2 + 48*x + 28
            sage: f[:50] == f
            True
        """
        return self._parent._base(zz_p_rep(zz_pX_GetCoeff(self.x, n)))

    def _unsafe_mutate(self, n, value):
        self.c.restore_c()
        zz_pX_SetCoeff_long(self.x, n, value)

    cpdef _add_(self, _right):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: (x + 5) + (x^2 - 6)
            x^2 + x + 99
        """
        cdef Polynomial_dense_modn_ntl_zz right = <Polynomial_dense_modn_ntl_zz>_right
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef bint do_sig = zz_pX_deg(self.x) + zz_pX_deg(right.x) > 1000000
        if do_sig: sig_on()
        self.c.restore_c()
        zz_pX_add(r.x, self.x, right.x)
        if do_sig: sig_off()
        return r

    cpdef _sub_(self, _right):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: (x + 5) - (x^2 - 6)
            99*x^2 + x + 11
        """
        cdef Polynomial_dense_modn_ntl_zz right = <Polynomial_dense_modn_ntl_zz>_right
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef bint do_sig = zz_pX_deg(self.x) + zz_pX_deg(right.x) > 1000000
        if do_sig: sig_on()
        self.c.restore_c()
        zz_pX_sub(r.x, self.x, right.x)
        if do_sig: sig_off()
        return r

    cpdef _mul_(self, _right):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: (x + 5) * (x^2 - 1)
            x^3 + 5*x^2 + 99*x + 95
        """
        cdef Polynomial_dense_modn_ntl_zz right = <Polynomial_dense_modn_ntl_zz>_right
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef bint do_sig = zz_pX_deg(self.x) + zz_pX_deg(right.x) > 10000
        if do_sig: sig_on()
        self.c.restore_c()
        if self is right:
            zz_pX_sqr(r.x, self.x)
        else:
            zz_pX_mul(r.x, self.x, right.x)
        if do_sig: sig_off()
        return r

    cpdef Polynomial _mul_trunc_(self, Polynomial right, long n):
        r"""
        Return the product of ``self`` and ``right`` truncated to the
        given length `n`

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(100), implementation="NTL")
            sage: f = x - 2
            sage: g = x^2 - 8*x + 16
            sage: f*g
            x^3 + 90*x^2 + 32*x + 68
            sage: f._mul_trunc_(g, 42)
            x^3 + 90*x^2 + 32*x + 68
            sage: f._mul_trunc_(g, 3)
            90*x^2 + 32*x + 68
            sage: f._mul_trunc_(g, 2)
            32*x + 68
            sage: f._mul_trunc_(g, 1)
            68
            sage: f._mul_trunc_(g, 0)
            0
            sage: f = x^2 - 8*x + 16
            sage: f._mul_trunc_(f, 2)
            44*x + 56
        """
        cdef Polynomial_dense_modn_ntl_zz op2 = <Polynomial_dense_modn_ntl_zz> right
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef bint do_sig = zz_pX_deg(self.x) + zz_pX_deg(op2.x) > 10000
        if do_sig: sig_on()
        self.c.restore_c()
        if self is op2:
            zz_pX_SqrTrunc(r.x, self.x, n)
        else:
            zz_pX_MulTrunc(r.x, self.x, op2.x, n)
        if do_sig: sig_off()
        return r

    cpdef _rmul_(self, Element c):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: (x+5) * 3
            3*x + 15
        """
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef bint do_sig = zz_pX_deg(self.x) > 100000
        if do_sig: sig_on()
        self.c.restore_c()
        zz_pX_rmul(r.x, self.x, c)
        if do_sig: sig_off()
        return r

    cpdef _lmul_(self, Element c):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: 3 * (x+5)
            3*x + 15
        """
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef bint do_sig = zz_pX_deg(self.x) > 100000
        if do_sig: sig_on()
        self.c.restore_c()
        zz_pX_lmul(r.x, c, self.x)
        if do_sig: sig_off()
        return r

    def __pow__(Polynomial_dense_modn_ntl_zz self, ee, modulus):
        r"""
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: (x-1)^5
            x^5 + 95*x^4 + 10*x^3 + 90*x^2 + 5*x + 99

        We define ``0^0`` to be unity, :issue:`13895`::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: R(0)^0
            1

        The value returned from ``0^0`` should belong to our ring::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: type(R(0)^0) == type(R(0))
            True

        Negative powers work (over prime fields) but use the generic
        implementation of fraction fields::

            sage: R.<x> = PolynomialRing(Integers(101), implementation='NTL')
            sage: f = (x-1)^(-5)
            sage: type(f)
            <class 'sage.rings.fraction_field_element.FractionFieldElement_1poly_field'>
            sage: (f + 2).numerator()
            2*x^5 + 91*x^4 + 20*x^3 + 81*x^2 + 10*x + 100

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: (x-1)^(-5)
            Traceback (most recent call last):
            ...
            TypeError: ...
        """
        cdef bint recip = 0, do_sig
        cdef long e = ee
        if e != ee:
            raise TypeError("Only integral powers defined.")
        elif e < 0:
            recip = 1
            e = -e
        if self == 0 and e == 0:
            return self.parent(1)

        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef zz_pX_Modulus_c mod

        self.c.restore_c()

        if modulus is None:
            if zz_pX_IsX(self.x):
                zz_pX_LeftShift(r.x, self.x, e-1)
            else:
                do_sig = zz_pX_deg(self.x) *e > 1000
                if do_sig: sig_on()
                zz_pX_power(r.x, self.x, e)
                if do_sig: sig_off()
        else:
            if not isinstance(modulus, Polynomial_dense_modn_ntl_zz):
                modulus = self.parent().coerce(modulus)
            zz_pX_Modulus_build(mod, (<Polynomial_dense_modn_ntl_zz>modulus).x)

            do_sig = zz_pX_deg(self.x) * e * self.c.p_bits > 1e5
            if do_sig: sig_on()
            zz_pX_PowerMod_long_pre(r.x, self.x, e, mod)
            if do_sig: sig_off()

        if recip:
            return ~r
        else:
            return r

    @coerce_binop
    def quo_rem(self, right):
        r"""
        Return `q` and `r`, with the degree of `r` less than the degree of ``right``,
        such that `q \cdot` ``right`` `{}+ r =` ``self``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(125), implementation='NTL')
            sage: f = x^5+1; g = (x+1)^2
            sage: q, r = f.quo_rem(g)
            sage: q
            x^3 + 123*x^2 + 3*x + 121
            sage: r
            5*x + 5
            sage: q*g + r
            x^5 + 1
        """
        cdef Polynomial_dense_modn_ntl_zz q = self._new()
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        cdef Polynomial_dense_modn_ntl_zz denom = <Polynomial_dense_modn_ntl_zz>right
        sig_on()
        self.c.restore_c()
        zz_pX_divrem(q.x, r.x, self.x, denom.x)
        sig_off()
        return q, r

    cpdef _floordiv_(self, right):
        """
        Return the whole part of ``self``/``right``, without remainder.

        For q = n // d, we have deg(n - q*d) < deg(d)

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(25), implementation='NTL')
            sage: f = x^7 + 1; g = x^2 - 1
            sage: q = f // g; q
            x^5 + x^3 + x
            sage: f - q*g
            x + 1
        """
        cdef Polynomial_dense_modn_ntl_zz numer = <Polynomial_dense_modn_ntl_zz>self
        cdef Polynomial_dense_modn_ntl_zz denom = <Polynomial_dense_modn_ntl_zz>right
        cdef Polynomial_dense_modn_ntl_zz q = numer._new()
        sig_on()
        numer.c.restore_c()
        zz_pX_div(q.x, numer.x, denom.x)
        sig_off()
        return q

    cpdef _mod_(self, right):
        """
        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(81), implementation='NTL')
            sage: f = x^7 + x + 1; g = x^3
            sage: r = f % g; r
            x + 1
            sage: g * x^4 + r
            x^7 + x + 1
        """
        cdef Polynomial_dense_modn_ntl_zz numer = <Polynomial_dense_modn_ntl_zz>self
        cdef Polynomial_dense_modn_ntl_zz denom = <Polynomial_dense_modn_ntl_zz>right
        cdef Polynomial_dense_modn_ntl_zz r = numer._new()
        sig_on()
        numer.c.restore_c()
        zz_pX_mod(r.x, numer.x, denom.x)
        sig_off()
        return r

    def shift(self, n):
        """
        Shift ``self`` to left by `n`, which is multiplication by `x^n`,
        truncating if `n` is negative.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = x^7 + x + 1
            sage: f.shift(1)
            x^8 + x^2 + x
            sage: f.shift(-1)
            x^6 + 1
            sage: f.shift(10).shift(-10) == f
            True

        TESTS::

            sage: p = R(0)
            sage: p.shift(3).is_zero()
            True
            sage: p.shift(-3).is_zero()
            True
        """
        return self << n

    def __lshift__(Polynomial_dense_modn_ntl_zz self, long n):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = x^5 + 2*x + 1
            sage: f << 3
            x^8 + 2*x^4 + x^3
        """
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        zz_pX_LeftShift(r.x, self.x, n)
        return r

    def __rshift__(Polynomial_dense_modn_ntl_zz self, long n):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = x^5 + 2*x + 1
            sage: f >> 3
            x^2
        """
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        zz_pX_RightShift(r.x, self.x, n)
        return r

    def _derivative(self, var=None):
        r"""
        Return the formal derivative of ``self`` with respect to ``var``.

        ``var`` must be either the generator of the polynomial ring to which
        this polynomial belongs, or ``None`` (either way the behaviour is the
        same).

        .. SEEALSO:: :meth:`.derivative`

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = x^4 - x - 1
            sage: f._derivative()
            4*x^3 + 76
            sage: f._derivative(None)
            4*x^3 + 76

            sage: f._derivative(2*x)
            Traceback (most recent call last):
            ...
            ValueError: cannot differentiate with respect to 2*x

        TESTS::

            sage: y = var("y")                                                          # needs sage.symbolic
            sage: f._derivative(y)                                                      # needs sage.symbolic
            Traceback (most recent call last):
            ...
            ValueError: cannot differentiate with respect to y
        """
        if var is not None and var != self._parent.gen():
            raise ValueError('cannot differentiate with respect to {}'.format(var))

        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        zz_pX_diff(r.x, self.x)
        return r

    def reverse(self, degree=None):
        """
        Return the reverse of the input polynomial thought as a polynomial of
        degree ``degree``.

        If `f` is a degree-`d` polynomial, its reverse is `x^d f(1/x)`.

        INPUT:

        - ``degree`` -- ``None`` or integer; if specified, truncate or zero
          pad the list of coefficients to this degree before reversing it

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = x^4 - x - 1
            sage: f.reverse()
            76*x^4 + 76*x^3 + 1
            sage: f.reverse(2)
            76*x^2 + 76*x
            sage: f.reverse(5)
            76*x^5 + 76*x^4 + x
            sage: g = x^3 - x
            sage: g.reverse()
            76*x^2 + 1

        TESTS:

        We check that this implementation is compatible with the generic one::

            sage: all(f.reverse(d) == Polynomial.reverse(f, d)
            ....:     for d in [None, 0, 1, 2, 3, 4, 5])
            True
        """
        cdef Polynomial_dense_modn_ntl_zz r = self._new()

        if degree is None:
            zz_pX_reverse(r.x, self.x)
        else:
            zz_pX_reverse_hi(r.x, self.x, degree)
        return r

    def is_gen(self):
        return zz_pX_IsX(self.x)

    def __bool__(self):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = x^4 - x - 1
            sage: not f
            False
            sage: not (x-x)
            True
        """
        return not zz_pX_IsZero(self.x)

    def valuation(self):
        """
        Return the valuation of ``self``, that is, the power of the
        lowest nonzero monomial of ``self``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(10), implementation='NTL')
            sage: x.valuation()
            1
            sage: f = x-3; f.valuation()
            0
            sage: f = x^99; f.valuation()
            99
            sage: f = x-x; f.valuation()
            +Infinity
        """
        cdef long n
        for n from 0 <= n <= zz_pX_deg(self.x):
            if zz_p_rep(zz_pX_GetCoeff(self.x, n)):
                return n
        return infinity

    def degree(self):
        """
        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = x^4 - x - 1
            sage: f.degree()
            4
            sage: f = 77*x + 1
            sage: f.degree()
            0
        """
        return zz_pX_deg(self.x)

    cpdef Polynomial truncate(self, long n):
        """
        Return this polynomial mod `x^n`.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(77), implementation='NTL')
            sage: f = sum(x^n for n in range(10)); f
            x^9 + x^8 + x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + x + 1
            sage: f.truncate(6)
            x^5 + x^4 + x^3 + x^2 + x + 1
        """
        cdef Polynomial_dense_modn_ntl_zz r = self._new()
        zz_pX_trunc(r.x, self.x, n)
        return r

    def __call__(self, *args, **kwds):
        """
        Evaluate ``self`` at ``x``. If ``x`` is a single argument coercible into
        the base ring of ``self``, this is done directly in NTL, otherwise
        the generic Polynomial call code is used.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: f = x^3 + 7
            sage: f(5)
            32
            sage: f(5r)
            32
            sage: f(mod(5, 1000))
            32
            sage: f(x)
            x^3 + 7
            sage: S.<y> = PolynomialRing(Integers(5), implementation='NTL')
            sage: f(y)
            y^3 + 2

        TESTS::

            sage: R.<x> = PolynomialRing(Integers(100), implementation='NTL')
            sage: f = x^3 + 7
            sage: f(1).parent() == R.base_ring()
            True
            sage: f(int(1)).parent() == R.base_ring()
            True
            sage: f(x + 1).parent() == f.parent()
            True

            sage: R.<x> = PolynomialRing(Zmod(12), 'x', implementation='NTL')
            sage: u = Zmod(4)(3)
            sage: x(u).parent() == u.parent()
            True
        """
        if len(args) != 1 or len(kwds) != 0:
            return Polynomial.__call__(self, *args, **kwds)
        arg = args[0]
        cdef ntl_zz_p fx = ntl_zz_p(0, self.c), x = None
        if isinstance(arg, int):
            x = ntl_zz_p(arg, self.c)
        elif isinstance(arg, Integer):
            x = ntl_zz_p(arg, self.c)
        elif isinstance(arg, Element):
            if <void *>self._parent._base == <void *>(<Element>arg)._parent: # c++ pointer hack
                x = ntl_zz_p(arg, self.c)
            else:
                map = self._parent._base.coerce_map_from((<Element>arg)._parent)
                if map is not None:
                    x = ntl_zz_p(map(arg), self.c)
        if x is None:
            return Polynomial.__call__(self, *args, **kwds)
        else:
            zz_pX_eval(fx.x, self.x, x.x)
            return self._parent._base(int(fx))


cdef class Polynomial_dense_modn_ntl_ZZ(Polynomial_dense_mod_n):

    def __init__(self, parent, v=None, check=True, is_gen=False, construct=False):
        if isinstance(v, Polynomial):
            if (<Element>v)._parent == parent:
                Polynomial.__init__(self, parent, is_gen=is_gen)
                self.x = (<Polynomial_dense_modn_ntl_ZZ>v).x
                self.c = (<Polynomial_dense_modn_ntl_ZZ>v).c
                return

        Polynomial_dense_mod_n.__init__(self, parent, v, check=check, is_gen=is_gen, construct=construct)
        cdef ntl_ZZ_pX ntl = self._poly
        self._poly = None # this will eventually go away
        self.x = ntl.x
        self.c = ntl.c

    def __dealloc__(self):
        if <object>self.c is not None:
            self.c.restore_c()

    cdef Polynomial_dense_modn_ntl_ZZ _new(self):
        cdef Polynomial_dense_modn_ntl_ZZ y = <Polynomial_dense_modn_ntl_ZZ>Polynomial_dense_modn_ntl_ZZ.__new__(Polynomial_dense_modn_ntl_ZZ)
        y.c = self.c
        y._parent = self._parent
        return y

    cpdef list list(self, bint copy=True):
        return [self.get_unsafe(n) for n from 0 <= n <= self.degree()]

    cdef get_unsafe(self, Py_ssize_t n):
        """
        Return the `n`-th coefficient of ``self``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: from sage.rings.polynomial.polynomial_modn_dense_ntl import Polynomial_dense_modn_ntl_ZZ
            sage: f = Polynomial_dense_modn_ntl_ZZ(R,[2,1])^7
            sage: f[3]
            560
            sage: f[:6]
            84*x^5 + 280*x^4 + 560*x^3 + 672*x^2 + 448*x + 128
            sage: f[:50] == f
            True
        """
        self.c.restore_c()
        # TODO, make this faster
        cdef ntl_ZZ_p ntl = ntl_ZZ_p(0, self.c)
        ntl.x = ZZ_pX_coeff(self.x, n)
        return self._parent._base(ntl._integer_())

    def _unsafe_mutate(self, n, value):
        self.c.restore_c()
        cdef Integer a
        if isinstance(value, Integer):
            a = <Integer>value
        else:
            a = ZZ(value)
        cdef ntl_ZZ_p val = ntl_ZZ_p(a, self.c)
        ZZ_pX_SetCoeff(self.x, n, val.x)

    cpdef _add_(self, _right):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: (x + 5) + (x^2 - 6)
            x^2 + x + 999999999999999999999999999999
        """
        cdef Polynomial_dense_modn_ntl_ZZ right = <Polynomial_dense_modn_ntl_ZZ>_right
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        cdef bint do_sig = (ZZ_pX_deg(self.x) + ZZ_pX_deg(right.x)) * self.c.p_bits > 1e7
        if do_sig: sig_on()
        self.c.restore_c()
        ZZ_pX_add(r.x, self.x, right.x)
        if do_sig: sig_off()
        return r

    cpdef _sub_(self, _right):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: (x + 5) - (x^2 - 6)
            999999999999999999999999999999*x^2 + x + 11
        """
        cdef Polynomial_dense_modn_ntl_ZZ right = <Polynomial_dense_modn_ntl_ZZ>_right
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        cdef bint do_sig = (ZZ_pX_deg(self.x) + ZZ_pX_deg(right.x)) * self.c.p_bits > 1e7
        if do_sig: sig_on()
        self.c.restore_c()
        ZZ_pX_sub(r.x, self.x, right.x)
        if do_sig: sig_off()
        return r

    cpdef _mul_(self, _right):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: (x+5) * (x^2 - 1)
            x^3 + 5*x^2 + 999999999999999999999999999999*x + 999999999999999999999999999995
        """
        cdef Polynomial_dense_modn_ntl_ZZ right = <Polynomial_dense_modn_ntl_ZZ>_right
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        cdef bint do_sig = (ZZ_pX_deg(self.x) + ZZ_pX_deg(right.x)) * self.c.p_bits > 1e5
        if do_sig: sig_on()
        self.c.restore_c()
        if self is right:
            ZZ_pX_sqr(r.x, self.x)
        else:
            ZZ_pX_mul(r.x, self.x, right.x)
        if do_sig: sig_off()
        return r

    cpdef Polynomial _mul_trunc_(self, Polynomial right, long n):
        """
        Return the product of ``self`` and ``right`` truncated to the
        given length `n`, only return terms of degree less than `n`.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation="NTL")
            sage: f = x - 2
            sage: g = x^2 - 8*x + 16
            sage: f*g
            x^3 + 999999999999999999999999999990*x^2 + 32*x + 999999999999999999999999999968
            sage: f._mul_trunc_(g, 42)
            x^3 + 999999999999999999999999999990*x^2 + 32*x + 999999999999999999999999999968
            sage: f._mul_trunc_(g, 3)
            999999999999999999999999999990*x^2 + 32*x + 999999999999999999999999999968
            sage: f._mul_trunc_(g, 2)
            32*x + 999999999999999999999999999968
            sage: f._mul_trunc_(g, 1)
            999999999999999999999999999968
            sage: f._mul_trunc_(g, 0)
            0
            sage: f = x^2 - 8*x + 16
            sage: f._mul_trunc_(f, 2)
            999999999999999999999999999744*x + 256
        """
        cdef Polynomial_dense_modn_ntl_ZZ op2 = <Polynomial_dense_modn_ntl_ZZ> right
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        cdef bint do_sig = (ZZ_pX_deg(self.x) + ZZ_pX_deg(op2.x)) * self.c.p_bits > 1e5
        if do_sig: sig_on()
        self.c.restore_c()
        if self is op2:
            ZZ_pX_SqrTrunc(r.x, self.x, n)
        else:
            ZZ_pX_MulTrunc(r.x, self.x, op2.x, n)
        if do_sig: sig_off()
        return r

    cpdef _rmul_(self, Element c):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: (x+5) * 3
            3*x + 15
        """
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        cdef bint do_sig = ZZ_pX_deg(self.x) * self.c.p_bits > 1e7
        if do_sig: sig_on()
        self.c.restore_c()
        cdef ntl_ZZ_p value = ntl_ZZ_p(c, self.c)
        ZZ_pX_rmul(r.x, self.x, value.x)
        if do_sig: sig_off()
        return r

    cpdef _lmul_(self, Element c):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: 3 * (x+5)
            3*x + 15
        """
        return self._rmul_(c)

    def __pow__(Polynomial_dense_modn_ntl_ZZ self, ee, modulus):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: (x+1)^5
            x^5 + 5*x^4 + 10*x^3 + 10*x^2 + 5*x + 1

        We define ``0^0`` to be unity, :issue:`13895`::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: R(0)^0
            1

        The value returned from ``0^0`` should belong to our ring::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: type(R(0)^0) == type(R(0))
            True
        """
        cdef bint recip = 0, do_sig
        cdef long e = ee
        if e != ee:
            raise TypeError("Only integral powers defined.")
        elif e < 0:
            recip = 1 # delay because powering frac field elements is slow
            e = -e
        if self == 0 and e == 0:
            return self.parent(1)
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        cdef ZZ_pX_Modulus_c mod

        self.c.restore_c()

        if modulus is None:
            if ZZ_pX_IsX(self.x):
                ZZ_pX_LeftShift(r.x, self.x, e - 1)
            else:
                do_sig = ZZ_pX_deg(self.x) * e * self.c.p_bits > 1e5
                if do_sig: sig_on()
                ZZ_pX_power(r.x, self.x, e)
                if do_sig: sig_off()
        else:
            if not isinstance(modulus, Polynomial_dense_modn_ntl_ZZ):
                modulus = self.parent().coerce(modulus)
            ZZ_pX_Modulus_build(mod, (<Polynomial_dense_modn_ntl_ZZ>modulus).x)

            do_sig = ZZ_pX_deg(self.x) * e * self.c.p_bits > 1e5
            if do_sig: sig_on()
            ZZ_pX_PowerMod_long_pre(r.x, self.x, e, mod)
            if do_sig: sig_off()
        if recip:
            return ~r
        else:
            return r

    @coerce_binop
    def quo_rem(self, right):
        r"""
        Return `q` and `r`, with the degree of `r` less than the degree of ``right``,
        such that `q \cdot` ``right`` `+ r =` ``self``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: f = x^5+1; g = (x+1)^2
            sage: q, r = f.quo_rem(g)
            sage: q
            x^3 + 999999999999999999999999999998*x^2 + 3*x + 999999999999999999999999999996
            sage: r
            5*x + 5
            sage: q*g + r
            x^5 + 1
        """
        cdef Polynomial_dense_modn_ntl_ZZ q = self._new()
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        cdef Polynomial_dense_modn_ntl_ZZ denom = <Polynomial_dense_modn_ntl_ZZ>right
        sig_on()
        self.c.restore_c()
        ZZ_pX_DivRem(q.x, r.x, self.x, denom.x)
        sig_off()
        return q, r

    cpdef _floordiv_(self, right):
        """
        Return the whole part of ``self`` / ``right``, without remainder.

        For q = n // d, we have deg(n - q*d) < deg(d)

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: f = x^7 + 1; g = x^2 - 1
            sage: q = f // g; q
            x^5 + x^3 + x
            sage: f - q*g
            x + 1
        """
        cdef Polynomial_dense_modn_ntl_ZZ numer = <Polynomial_dense_modn_ntl_ZZ>self
        cdef Polynomial_dense_modn_ntl_ZZ denom = <Polynomial_dense_modn_ntl_ZZ>right
        cdef Polynomial_dense_modn_ntl_ZZ q = numer._new()
        sig_on()
        numer.c.restore_c()
        ZZ_pX_div(q.x, numer.x, denom.x)
        sig_off()
        return q

    cpdef _mod_(self, right):
        """
        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(9^30), implementation='NTL')
            sage: f = x^7 + x + 1; g = x^3 - 1
            sage: r = f % g; r
            2*x + 1
            sage: g * (x^4 + x) + r
            x^7 + x + 1
        """
        cdef Polynomial_dense_modn_ntl_ZZ numer = <Polynomial_dense_modn_ntl_ZZ>self
        cdef Polynomial_dense_modn_ntl_ZZ denom = <Polynomial_dense_modn_ntl_ZZ>right
        cdef Polynomial_dense_modn_ntl_ZZ r = numer._new()
        sig_on()
        numer.c.restore_c()
        ZZ_pX_rem(r.x, numer.x, denom.x)
        sig_off()
        return r

    def shift(self, n):
        """
        Shift ``self`` to left by `n`, which is multiplication by `x^n`,
        truncating if `n` is negative.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(12^30), implementation='NTL')
            sage: f = x^7 + x + 1
            sage: f.shift(1)
            x^8 + x^2 + x
            sage: f.shift(-1)
            x^6 + 1
            sage: f.shift(10).shift(-10) == f
            True

        TESTS::

            sage: p = R(0)
            sage: p.shift(3).is_zero()
            True
            sage: p.shift(-3).is_zero()
            True
        """
        return self << n

    def __lshift__(Polynomial_dense_modn_ntl_ZZ self, long n):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(14^30), implementation='NTL')
            sage: f = x^5 + 2*x + 1
            sage: f << 3
            x^8 + 2*x^4 + x^3
        """
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        ZZ_pX_LeftShift(r.x, self.x, n)
        return r

    def __rshift__(Polynomial_dense_modn_ntl_ZZ self, long n):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(15^30), implementation='NTL')
            sage: f = x^5 + 2*x + 1
            sage: f >> 3
            x^2
        """
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        ZZ_pX_RightShift(r.x, self.x, n)
        return r

    def _derivative(self, var=None):
        r"""
        Return the formal derivative of ``self`` with respect to ``var``.

        ``var`` must be either the generator of the polynomial ring to which
        this polynomial belongs, or None (either way the behaviour is the
        same).

        .. SEEALSO:: :meth:`.derivative`

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(12^29), implementation='NTL')
            sage: f = x^4 + x + 5
            sage: f._derivative()
            4*x^3 + 1
            sage: f._derivative(None)
            4*x^3 + 1

            sage: f._derivative(2*x)
            Traceback (most recent call last):
            ...
            ValueError: cannot differentiate with respect to 2*x

        TESTS::

            sage: y = var("y")                                                          # needs sage.symbolic
            sage: f._derivative(y)                                                      # needs sage.symbolic
            Traceback (most recent call last):
            ...
            ValueError: cannot differentiate with respect to y
        """
        if var is not None and var != self._parent.gen():
            raise ValueError("cannot differentiate with respect to {}".format(var))

        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        ZZ_pX_diff(r.x, self.x)
        return r

    def reverse(self, degree=None):
        """
        Return the reverse of the input polynomial thought as a polynomial of
        degree ``degree``.

        If `f` is a degree-`d` polynomial, its reverse is `x^d f(1/x)`.

        INPUT:

        - ``degree`` -- ``None`` or integer; if specified, truncate or zero
          pad the list of coefficients to this degree before reversing it

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(12^29), implementation='NTL')
            sage: f = x^4 + 2*x + 5
            sage: f.reverse()
            5*x^4 + 2*x^3 + 1
            sage: f = x^3 + x
            sage: f.reverse()
            x^2 + 1
            sage: f.reverse(1)
            1
            sage: f.reverse(5)
            x^4 + x^2

        TESTS:

        We check that this implementation is compatible with the generic one::

            sage: all(f.reverse(d) == Polynomial.reverse(f, d)
            ....:     for d in [None, 0, 1, 2, 3, 4, 5])
            True
        """
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()

        if degree is None:
            ZZ_pX_reverse(r.x, self.x)
        else:
            ZZ_pX_reverse_hi(r.x, self.x, degree)
        return r

    def is_gen(self):
        return ZZ_pX_IsX(self.x)

    def valuation(self):
        """
        Return the valuation of ``self``, that is, the power of the
        lowest nonzero monomial of ``self``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(10^50), implementation='NTL')
            sage: x.valuation()
            1
            sage: f = x - 3; f.valuation()
            0
            sage: f = x^99; f.valuation()
            99
            sage: f = x - x; f.valuation()
            +Infinity
        """
        cdef long n
        cdef ZZ_p_c coeff
        for n from 0 <= n <= ZZ_pX_deg(self.x):
            coeff = ZZ_pX_coeff(self.x, n)
            if not ZZ_p_IsZero(coeff):
                return n
        return infinity

    def __bool__(self):
        """
        TESTS::

            sage: R.<x> = PolynomialRing(Integers(12^29), implementation='NTL')
            sage: f = x^4 + 1
            sage: not f
            False
            sage: not (x-x)
            True
        """
        return not ZZ_pX_IsZero(self.x)

    def degree(self):
        """
        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(14^34), implementation='NTL')
            sage: f = x^4 - x - 1
            sage: f.degree()
            4
            sage: f = 14^43*x + 1
            sage: f.degree()
            0
        """
        return ZZ_pX_deg(self.x)

    cpdef Polynomial truncate(self, long n):
        """
        Return this polynomial mod `x^n`.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(15^30), implementation='NTL')
            sage: f = sum(x^n for n in range(10)); f
            x^9 + x^8 + x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + x + 1
            sage: f.truncate(6)
            x^5 + x^4 + x^3 + x^2 + x + 1
        """
        cdef Polynomial_dense_modn_ntl_ZZ r = self._new()
        ZZ_pX_trunc(r.x, self.x, n)
        return r

    def __call__(self, *args, **kwds):
        """
        Evaluate ``self`` at ``x``. If ``x`` is a single argument coercible into
        the base ring of ``self``, this is done directly in NTL, otherwise
        the generic Polynomial call code is used.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: f = x^3 + 7
            sage: f(5)
            132
            sage: f(5r)
            132
            sage: f(mod(5, 10^50))
            132
            sage: f(x)
            x^3 + 7
            sage: S.<y> = PolynomialRing(Integers(5), implementation='NTL')
            sage: f(y)
            y^3 + 2

        TESTS::

            sage: R.<x> = PolynomialRing(Integers(10^30), implementation='NTL')
            sage: f = x^3 + 7
            sage: f(1).parent() == R.base_ring()
            True
            sage: f(int(1)).parent() == R.base_ring()
            True
            sage: f(x + 1).parent() == f.parent()
            True

            sage: R.<x> = PolynomialRing(Zmod(10^30), 'x', implementation='NTL')
            sage: u = Zmod(10^29)(3)
            sage: x(u).parent() == u.parent()
            True
            sage: v = Zmod(10)(3)
            sage: x(v).parent() == v.parent()
            True
        """
        if len(args) != 1 or len(kwds) != 0:
            return Polynomial.__call__(self, *args, **kwds)
        arg = args[0]
        cdef ntl_ZZ_p fx = ntl_ZZ_p(0, self.c), x = None
        if isinstance(arg, (int, Integer)):
            x = ntl_ZZ_p(arg, self.c)
        elif isinstance(arg, Element):
            if <void *>self._parent._base == <void *>(<Element>arg)._parent: # c++ pointer hack
                x = ntl_ZZ_p(arg, self.c)
            else:
                map = self._parent._base.coerce_map_from((<Element>arg)._parent)
                if map is not None:
                    x = ntl_ZZ_p(map(arg), self.c)
        if x is None:
            return Polynomial.__call__(self, *args, **kwds)
        else:
            ZZ_pX_eval(fx.x, self.x, x.x)
            return self._parent._base(fx._integer_())


cdef class Polynomial_dense_mod_p(Polynomial_dense_mod_n):
    """
    A dense polynomial over the integers modulo `p`, where `p` is prime.
    """

    def __pow__(self, n, modulus):
        """
        Exponentiation of ``self``.

        If ``modulus`` is not ``None``, the exponentiation is performed
        modulo the polynomial ``modulus``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(Integers(101), implementation='NTL')
            sage: (x-1)^5
            x^5 + 96*x^4 + 10*x^3 + 91*x^2 + 5*x + 100
            sage: pow(x-1, 15, x^3+x+1)
            55*x^2 + 6*x + 46

        TESTS:

        Negative powers work but use the generic
        implementation of fraction fields::

            sage: R.<x> = PolynomialRing(Integers(101), implementation='NTL')
            sage: f = (x-1)^(-5)
            sage: type(f)
            <class 'sage.rings.fraction_field_element.FractionFieldElement_1poly_field'>
            sage: (f + 2).numerator()
            2*x^5 + 91*x^4 + 20*x^3 + 81*x^2 + 10*x + 100

        We define ``0^0`` to be unity, :issue:`13895`::

            sage: R.<x> = PolynomialRing(Integers(101), implementation='NTL')
            sage: R(0)^0
            1

        The value returned from ``0^0`` should belong to our ring::

            sage: R.<x> = PolynomialRing(Integers(101), implementation='NTL')
            sage: type(R(0)^0) == type(R(0))
            True

        The modulus can have smaller degree than ``self``::

            sage: R.<x> = PolynomialRing(GF(101), implementation="NTL")
            sage: pow(x^4 + 1, 100, x^2 + x + 1)
            100*x + 100

        Canonical coercion should apply::

            sage: # needs sage.libs.flint
            sage: R.<x> = PolynomialRing(GF(101), implementation="FLINT")
            sage: x_ZZ = ZZ["x"].gen()
            sage: pow(x+1, 100, 2)
            0
            sage: pow(x+1, 100, x_ZZ^2 + x_ZZ + 1)
            100*x + 100
            sage: pow(x+1, int(100), x_ZZ^2 + x_ZZ + 1)
            100*x + 100
            sage: xx = polygen(GF(97))
            sage: _ = pow(x + 1, 3, xx^3 + xx + 1)
            Traceback (most recent call last):
            ...
            TypeError: no common canonical parent for objects with parents: ...
        """
        n = Integer(n)
        parent = (<Element>self)._parent

        if modulus is not None:
            # Similar to coerce_binop
            if not have_same_parent(self, modulus):
                a, m = canonical_coercion(self, modulus)
                if a is not self:
                    return pow(a, n, m)
                modulus = m
            self = self % modulus
            return parent(pow(self.ntl_ZZ_pX(), n, modulus.ntl_ZZ_pX()), construct=True)
        else:
            if n < 0:
                return ~(self**(-n))
            elif self.degree() <= 0:
                return parent(self[0]**n)
            else:
                return parent(self.ntl_ZZ_pX()**n, construct=True)

    @coerce_binop
    def gcd(self, right):
        """
        Return the greatest common divisor of this polynomial and ``other``, as
        a monic polynomial.

        INPUT:

        - ``other`` -- a polynomial defined over the same ring as ``self``

        EXAMPLES::

            sage: R.<x> = PolynomialRing(GF(3), implementation="NTL")
            sage: f, g = x + 2, x^2 - 1
            sage: f.gcd(g)
            x + 2
        """
        g = self.ntl_ZZ_pX().gcd(right.ntl_ZZ_pX())
        return self.parent()(g, construct=True)

    @coerce_binop
    def xgcd(self, other):
        r"""
        Compute the extended gcd of this element and ``other``.

        INPUT:

        - ``other`` -- an element in the same polynomial ring

        OUTPUT:

        A tuple ``(r,s,t)`` of elements in the polynomial ring such
        that ``r = s*self + t*other``.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(GF(3), implementation='NTL')
            sage: x.xgcd(x)
            (x, 0, 1)
            sage: (x^2 - 1).xgcd(x - 1)
            (x + 2, 0, 1)
            sage: R.zero().xgcd(R.one())
            (1, 0, 1)
            sage: (x^3 - 1).xgcd((x - 1)^2)
            (x^2 + x + 1, 0, 1)
            sage: ((x - 1)*(x + 1)).xgcd(x*(x - 1))
            (x + 2, 1, 2)
        """
        r, s, t = self.ntl_ZZ_pX().xgcd(other.ntl_ZZ_pX())
        return self.parent()(r, construct=True), self.parent()(s, construct=True), self.parent()(t, construct=True)

    @coerce_binop
    def resultant(self, other):
        """
        Return the resultant of ``self`` and ``other``, which must lie in the same
        polynomial ring.

        INPUT:

        - ``other`` -- a polynomial

        OUTPUT: an element of the base ring of the polynomial ring

        EXAMPLES::

            sage: R.<x> = PolynomialRing(GF(19), implementation='NTL')
            sage: f = x^3 + x + 1;  g = x^3 - x - 1
            sage: r = f.resultant(g); r
            11
            sage: r.parent() is GF(19)
            True
        """
        other = self.parent().coerce(other)
        return self.base_ring()(str(self.ntl_ZZ_pX().resultant(other.ntl_ZZ_pX())))

    def discriminant(self):
        """
        EXAMPLES::

            sage: _.<x> = PolynomialRing(GF(19), implementation='NTL')
            sage: f = x^3 + 3*x - 17
            sage: f.discriminant()
            12
        """
        return self.base_ring()(str(self.ntl_ZZ_pX().discriminant()))



# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.libs.pari
r"""
Ambient spaces of modular forms

EXAMPLES:

We compute a basis for the ambient space
`M_2(\Gamma_1(25),\chi)`, where `\chi` is
quadratic.

::

    sage: chi = DirichletGroup(25,QQ).0; chi
    Dirichlet character modulo 25 of conductor 5 mapping 2 |--> -1
    sage: n = ModularForms(chi,2); n
    Modular Forms space of dimension 6, character [-1] and weight 2 over Rational Field
    sage: type(n)
    <class 'sage.modular.modform.ambient_eps.ModularFormsAmbient_eps_with_category'>

Compute a basis::

    sage: n.basis()
    [1 + O(q^6),
     q + O(q^6),
     q^2 + O(q^6),
     q^3 + O(q^6),
     q^4 + O(q^6),
     q^5 + O(q^6)]

Compute the same basis but to higher precision::

    sage: n.set_precision(20)
    sage: n.basis()
    [1 + 10*q^10 + 20*q^15 + O(q^20),
     q + 5*q^6 + q^9 + 12*q^11 - 3*q^14 + 17*q^16 + 8*q^19 + O(q^20),
     q^2 + 4*q^7 - q^8 + 8*q^12 + 2*q^13 + 10*q^17 - 5*q^18 + O(q^20),
     q^3 + q^7 + 3*q^8 - q^12 + 5*q^13 + 3*q^17 + 6*q^18 + O(q^20),
     q^4 - q^6 + 2*q^9 + 3*q^14 - 2*q^16 + 4*q^19 + O(q^20),
     q^5 + q^10 + 2*q^15 + O(q^20)]

TESTS::

    sage: m = ModularForms(Gamma1(20),2,GF(7))
    sage: loads(dumps(m)) == m
    True

::

    sage: m = ModularForms(GammaH(11,[3]), 2); m
    Modular Forms space of dimension 2 for Congruence Subgroup Gamma_H(11) with H generated by [3] of weight 2 over Rational Field
    sage: type(m)
    <class 'sage.modular.modform.ambient_g1.ModularFormsAmbient_gH_Q_with_category'>
    sage: m == loads(dumps(m))
    True
"""

# ****************************************************************************
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import is_prime, sigma
from sage.matrix.constructor import matrix
from sage.misc.cachefunc import cached_method
from sage.modular.arithgroup.congroup_gamma0 import Gamma0_class
from sage.modular.arithgroup.congroup_gamma1 import Gamma1_class
from sage.modular.arithgroup.congroup_generic import CongruenceSubgroupBase
from sage.modular.dirichlet import TrivialCharacter
from sage.modular.hecke.ambient_module import AmbientHeckeModule
from sage.modular.modsym.modsym import ModularSymbols
from sage.modules.free_module import VectorSpace
from sage.rings.integer import Integer
from sage.structure.sequence import Sequence

from . import defaults
from . import eisenstein_submodule
from . import eis_series
from . import space
from . import submodule


class ModularFormsAmbient(space.ModularFormsSpace,
                          AmbientHeckeModule):
    """
    An ambient space of modular forms.
    """
    def __init__(self, group, weight, base_ring, character=None, eis_only=False):
        """
        Create an ambient space of modular forms.

        EXAMPLES::

            sage: m = ModularForms(Gamma1(20),20); m
            Modular Forms space of dimension 238 for Congruence Subgroup Gamma1(20) of weight 20 over Rational Field
            sage: m.is_ambient()
            True
        """
        if not isinstance(group, CongruenceSubgroupBase):
            raise TypeError('group (=%s) must be a congruence subgroup' % group)
        weight = Integer(weight)

        if character is None and isinstance(group, Gamma0_class):
            character = TrivialCharacter(group.level(), base_ring)

        self._eis_only = eis_only
        space.ModularFormsSpace.__init__(self, group, weight, character, base_ring)
        if eis_only:
            d = self._dim_eisenstein()
        else:
            d = self.dimension()
        AmbientHeckeModule.__init__(self, base_ring, d, group.level(), weight)

    def _repr_(self):
        """
        Return string representation of ``self``.

        EXAMPLES::

            sage: m = ModularForms(Gamma1(20),100); m._repr_()
            'Modular Forms space of dimension 1198 for Congruence Subgroup Gamma1(20) of weight 100 over Rational Field'

        The output of _repr_ is not affected by renaming the space::

            sage: m.rename('A big modform space')
            sage: m
            A big modform space
            sage: m._repr_()
            'Modular Forms space of dimension 1198 for Congruence Subgroup Gamma1(20) of weight 100 over Rational Field'
        """
        if self._eis_only:
            return "Modular Forms space for %s of weight %s over %s" % (
                self.group(), self.weight(), self.base_ring())
        else:
            return "Modular Forms space of dimension %s for %s of weight %s over %s" % (
                self.dimension(), self.group(), self.weight(), self.base_ring())

    def _submodule_class(self):
        """
        Return the Python class of submodules of this modular forms space.

        EXAMPLES::

            sage: m = ModularForms(Gamma0(20),2)
            sage: m._submodule_class()
            <class 'sage.modular.modform.submodule.ModularFormsSubmodule'>
        """
        return submodule.ModularFormsSubmodule

    def change_ring(self, base_ring):
        """
        Change the base ring of this space of modular forms.

        INPUT:

        - ``R`` -- ring

        EXAMPLES::

            sage: M = ModularForms(Gamma0(37),2)
            sage: M.basis()
            [q + q^3 - 2*q^4 + O(q^6),
             q^2 + 2*q^3 - 2*q^4 + q^5 + O(q^6),
             1 + 2/3*q + 2*q^2 + 8/3*q^3 + 14/3*q^4 + 4*q^5 + O(q^6)]

        The basis after changing the base ring is the reduction modulo
        `3` of an integral basis.

        ::

            sage: M3 = M.change_ring(GF(3))
            sage: M3.basis()
            [q + q^3 + q^4 + O(q^6),
             q^2 + 2*q^3 + q^4 + q^5 + O(q^6),
             1 + q^3 + q^4 + 2*q^5 + O(q^6)]
        """
        from . import constructor
        return constructor.ModularForms(self.group(), self.weight(),
                                        base_ring, prec=self.prec(),
                                        eis_only=self._eis_only)

    @cached_method
    def dimension(self):
        """
        Return the dimension of this ambient space of modular forms,
        computed using a dimension formula (so it should be reasonably
        fast).

        EXAMPLES::

            sage: m = ModularForms(Gamma1(20),20)
            sage: m.dimension()
            238
        """
        return self._dim_eisenstein() + self._dim_cuspidal()

    def hecke_module_of_level(self, N):
        r"""
        Return the Hecke module of level N corresponding to self, which is the
        domain or codomain of a degeneracy map from ``self``. Here N must be either
        a divisor or a multiple of the level of ``self``.

        EXAMPLES::

            sage: ModularForms(25, 6).hecke_module_of_level(5)
            Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(5) of weight 6 over Rational Field
            sage: ModularForms(Gamma1(4), 3).hecke_module_of_level(8)
            Modular Forms space of dimension 7 for Congruence Subgroup Gamma1(8) of weight 3 over Rational Field
            sage: ModularForms(Gamma1(4), 3).hecke_module_of_level(9)
            Traceback (most recent call last):
            ...
            ValueError: N (=9) must be a divisor or a multiple of the level of self (=4)
        """
        if not (N % self.level() == 0 or self.level() % N == 0):
            raise ValueError("N (=%s) must be a divisor or a multiple of the level of self (=%s)" % (N, self.level()))
        from . import constructor
        return constructor.ModularForms(self.group()._new_group_from_level(N), self.weight(), self.base_ring(), prec=self.prec())

    def _degeneracy_raising_matrix(self, M, t):
        r"""
        Calculate the matrix of the degeneracy map from ``self`` to M corresponding
        to `f(q) \mapsto f(q^t)`. Here the level of M should be a multiple of
        the level of self, and t should divide the quotient.

        EXAMPLES::

            sage: ModularForms(22, 2)._degeneracy_raising_matrix(ModularForms(44, 2), 1)
            [  1   0  -1  -2   0   0   0   0   0]
            [  0   1   0  -2   0   0   0   0   0]
            [  0   0   0   0   1   0   0   0  24]
            [  0   0   0   0   0   1   0  -2  21]
            [  0   0   0   0   0   0   1   3 -10]
            sage: ModularForms(22, 2)._degeneracy_raising_matrix(ModularForms(44, 2), 2)
            [0 1 0 0 0 0 0 0 0]
            [0 0 0 1 0 0 0 0 0]
            [0 0 0 0 1 0 0 0 0]
            [0 0 0 0 0 0 1 0 0]
            [0 0 0 0 0 0 0 1 0]
        """
        from sage.matrix.matrix_space import MatrixSpace
        A = MatrixSpace(self.base_ring(), self.dimension(), M.dimension())
        d = M.sturm_bound() + 1
        q = self.an_element().qexp(d).parent().gen()
        im_gens = []
        for x in self.basis():
            fq = x.qexp(d)
            fqt = fq(q**t).add_bigoh(d)  # silly workaround for trac #5367
            im_gens.append(M(fqt))
        return A([M.coordinate_vector(u) for u in im_gens])

    def rank(self):
        r"""
        This is a synonym for ``self.dimension()``.

        EXAMPLES::

            sage: m = ModularForms(Gamma0(20),4)
            sage: m.rank()
            12
            sage: m.dimension()
            12
        """
        return self.dimension()

    def ambient_space(self):
        """
        Return the ambient space that contains this ambient space. This is,
        of course, just this space again.

        EXAMPLES::

            sage: m = ModularForms(Gamma0(3),30)
            sage: m.ambient_space() is m
            True
        """
        return self

    def is_ambient(self) -> bool:
        """
        Return ``True`` if this an ambient space of modular forms.

        This is an ambient space, so this function always returns ``True``.

        EXAMPLES::

            sage: ModularForms(11).is_ambient()
            True
            sage: CuspForms(11).is_ambient()
            False
        """
        return True

    @cached_method(key=lambda self, sign: Integer(sign))  # convert sign to an Integer before looking this up in the cache
    def modular_symbols(self, sign=0):
        """
        Return the corresponding space of modular symbols with the given
        sign.

        EXAMPLES::

            sage: S = ModularForms(11,2)
            sage: S.modular_symbols()
            Modular Symbols space of dimension 3 for Gamma_0(11) of weight 2 with sign 0 over Rational Field
            sage: S.modular_symbols(sign=1)
            Modular Symbols space of dimension 2 for Gamma_0(11) of weight 2 with sign 1 over Rational Field
            sage: S.modular_symbols(sign=-1)
            Modular Symbols space of dimension 1 for Gamma_0(11) of weight 2 with sign -1 over Rational Field

        ::

            sage: ModularForms(1,12).modular_symbols()
            Modular Symbols space of dimension 3 for Gamma_0(1) of weight 12 with sign 0 over Rational Field
        """
        sign = Integer(sign)
        return ModularSymbols(group=self.group(),
                                     weight=self.weight(),
                                     sign=sign,
                                     base_ring=self.base_ring())

    @cached_method
    def module(self):
        """
        Return the underlying free module corresponding to this space
        of modular forms.

        EXAMPLES::

            sage: m = ModularForms(Gamma1(13),10)
            sage: m.free_module()
            Vector space of dimension 69 over Rational Field
            sage: ModularForms(Gamma1(13),4, GF(49,'b')).free_module()
            Vector space of dimension 27 over Finite Field in b of size 7^2
        """
        d = self.dimension()
        return VectorSpace(self.base_ring(), d)

    # free_module -- stupid thing: there are functions in classes
    # ModularFormsSpace and HeckeModule that both do much the same
    # thing, and one has to override both of them!
    def free_module(self):
        """
        Return the free module underlying this space of modular forms.

        EXAMPLES::

            sage: ModularForms(37).free_module()
            Vector space of dimension 3 over Rational Field
        """
        return self.module()

    def prec(self, new_prec=None):
        """
        Set or get default initial precision for printing modular forms.

        INPUT:

        - ``new_prec`` -- positive integer (default: ``None``)

        OUTPUT: if ``new_prec`` is ``None``, returns the current precision

        EXAMPLES::

            sage: M = ModularForms(1,12, prec=3)
            sage: M.prec()
            3

        ::

            sage: M.basis()
            [q - 24*q^2 + O(q^3), 1 + 65520/691*q + 134250480/691*q^2 + O(q^3)]

        ::

            sage: M.prec(5)
            5
            sage: M.basis()
            [q - 24*q^2 + 252*q^3 - 1472*q^4 + O(q^5),
             1 + 65520/691*q + 134250480/691*q^2 + 11606736960/691*q^3 + 274945048560/691*q^4 + O(q^5)]
        """
        if new_prec:
            self.__prec = new_prec
        try:
            return self.__prec
        except AttributeError:
            self.__prec = defaults.DEFAULT_PRECISION
        return self.__prec

    def set_precision(self, n):
        """
        Set the default precision for displaying elements of this space.

        EXAMPLES::

            sage: m = ModularForms(Gamma1(5),2)
            sage: m.set_precision(10)
            sage: m.basis()
            [1 + 60*q^3 - 120*q^4 + 240*q^5 - 300*q^6 + 300*q^7 - 180*q^9 + O(q^10),
             q + 6*q^3 - 9*q^4 + 27*q^5 - 28*q^6 + 30*q^7 - 11*q^9 + O(q^10),
             q^2 - 4*q^3 + 12*q^4 - 22*q^5 + 30*q^6 - 24*q^7 + 5*q^8 + 18*q^9 + O(q^10)]
            sage: m.set_precision(5)
            sage: m.basis()
            [1 + 60*q^3 - 120*q^4 + O(q^5),
             q + 6*q^3 - 9*q^4 + O(q^5),
             q^2 - 4*q^3 + 12*q^4 + O(q^5)]
        """
        if n < 0:
            raise ValueError("n (=%s) must be >= 0" % n)
        self.__prec = Integer(n)

    ####################################################################
    # Computation of Special Submodules
    ####################################################################
    @cached_method
    def cuspidal_submodule(self):
        """
        Return the cuspidal submodule of this ambient module.

        EXAMPLES::

            sage: ModularForms(Gamma1(13)).cuspidal_submodule()
            Cuspidal subspace of dimension 2 of Modular Forms space of dimension 13 for
            Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
        """
        from .cuspidal_submodule import CuspidalSubmodule
        return CuspidalSubmodule(self)

    @cached_method
    def eisenstein_submodule(self):
        """
        Return the Eisenstein submodule of this ambient module.

        EXAMPLES::

            sage: m = ModularForms(Gamma1(13),2); m
            Modular Forms space of dimension 13 for Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
            sage: m.eisenstein_submodule()
            Eisenstein subspace of dimension 11 of Modular Forms space of dimension 13 for Congruence Subgroup Gamma1(13) of weight 2 over Rational Field
        """
        return eisenstein_submodule.EisensteinSubmodule(self)

    @cached_method(key=lambda self, p: (Integer(p) if p is not None else p))  # convert p to an Integer before looking this up in the cache
    def new_submodule(self, p=None):
        """
        Return the new or `p`-new submodule of this ambient
        module.

        INPUT:

        - ``p`` -- (default: ``None``), if specified return only
          the `p`-new submodule

        EXAMPLES::

            sage: m = ModularForms(Gamma0(33),2); m
            Modular Forms space of dimension 6 for Congruence Subgroup Gamma0(33) of weight 2 over Rational Field
            sage: m.new_submodule()
            Modular Forms subspace of dimension 1 of Modular Forms space of dimension 6 for Congruence Subgroup Gamma0(33) of weight 2 over Rational Field

        Another example::

            sage: M = ModularForms(17,4)
            sage: N = M.new_subspace(); N
            Modular Forms subspace of dimension 4 of Modular Forms space of dimension 6 for Congruence Subgroup Gamma0(17) of weight 4 over Rational Field
            sage: N.basis()
            [q + 2*q^5 + O(q^6),
             q^2 - 3/2*q^5 + O(q^6),
             q^3 + O(q^6),
             q^4 - 1/2*q^5 + O(q^6)]

        ::

            sage: ModularForms(12,4).new_submodule()
            Modular Forms subspace of dimension 1 of Modular Forms space of dimension 9 for Congruence Subgroup Gamma0(12) of weight 4 over Rational Field

        Unfortunately (TODO) - `p`-new submodules aren't yet
        implemented::

            sage: m.new_submodule(3)            # not implemented
            Traceback (most recent call last):
            ...
            NotImplementedError
            sage: m.new_submodule(11)           # not implemented
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        if p is not None:
            p = Integer(p)
            if not p.is_prime():
                raise ValueError("p (=%s) must be a prime or None." % p)
        return self.cuspidal_submodule().new_submodule(p) + self.eisenstein_submodule().new_submodule(p)

    def _q_expansion(self, element, prec):
        r"""
        Return the `q`-expansion of a particular element of this space of
        modular forms, where the element should be a vector, list, or tuple
        (not a ModularFormElement). Here element should have length =
        self.dimension(). If element = [ a_i ] and self.basis() = [ v_i
        ], then we return `\sum a_i v_i`.

        INPUT:

        - ``element`` -- vector, list or tuple

        - ``prec`` -- desired precision of `q`-expansion

        EXAMPLES::

            sage: m = ModularForms(Gamma0(23),2); m
            Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(23) of weight 2 over Rational Field
            sage: m.basis()
            [q - q^3 - q^4 + O(q^6),
             q^2 - 2*q^3 - q^4 + 2*q^5 + O(q^6),
             1 + 12/11*q + 36/11*q^2 + 48/11*q^3 + 84/11*q^4 + 72/11*q^5 + O(q^6)]
            sage: m._q_expansion([1,2,0], 5)
            q + 2*q^2 - 5*q^3 - 3*q^4 + O(q^5)
        """
        B = self.q_expansion_basis(prec)
        f = self._q_expansion_zero()
        for i in range(len(element)):
            if element[i]:
                f += element[i] * B[i]
        return f

    ####################################################################
    # Computations of Dimensions
    ####################################################################
    @cached_method
    def _dim_cuspidal(self):
        r"""
        Return the dimension of the cuspidal subspace of this ambient
        modular forms space.

        For weights `k \ge 2` this is computed using a
        dimension formula. For weight 1, it will trigger a computation of a
        basis of `q`-expansions using Schaeffer's algorithm, unless this space
        is a space of Eisenstein forms only, in which case we just return 0.

        EXAMPLES::

            sage: m = ModularForms(GammaH(11,[3]), 2); m
            Modular Forms space of dimension 2 for Congruence Subgroup Gamma_H(11) with H generated by [3] of weight 2 over Rational Field
            sage: m._dim_cuspidal()
            1
            sage: m = ModularForms(DirichletGroup(389,CyclotomicField(4)).0,3); m._dim_cuspidal()
            64
            sage: m = ModularForms(GammaH(31, [7]), 1)
            sage: m._dim_cuspidal()
            1
            sage: m = ModularForms(GammaH(31, [7]), 1, eis_only=True)
            sage: m._dim_cuspidal()
            0
        """
        if self._eis_only:
            return 0
        if isinstance(self.group(), Gamma1_class) and self.character() is not None:
            return self.group().dimension_cusp_forms(self.weight(),
                                                     self.character())
        else:
            return self.group().dimension_cusp_forms(self.weight())

    @cached_method
    def _dim_eisenstein(self):
        """
        Return the dimension of the Eisenstein subspace of this modular
        symbols space, computed using a dimension formula.

        EXAMPLES::

            sage: m = ModularForms(GammaH(13,[4]), 2); m
            Modular Forms space of dimension 3 for Congruence Subgroup Gamma_H(13) with H generated by [4] of weight 2 over Rational Field
            sage: m._dim_eisenstein()
            3

            sage: m = ModularForms(DirichletGroup(13).0,7); m
            Modular Forms space of dimension 8, character [zeta12] and weight 7 over Cyclotomic Field of order 12 and degree 4
            sage: m._dim_eisenstein()
            2
            sage: m._dim_cuspidal()
            6

        Test that :issue:`24030` is fixed::

            sage: ModularForms(GammaH(40, [21]), 1).dimension() # indirect doctest
            16
        """
        if isinstance(self.group(), Gamma1_class) and self.character() is not None:
            return self.group().dimension_eis(self.weight(), self.character())
        else:
            return self.group().dimension_eis(self.weight())

    @cached_method
    def _dim_new_cuspidal(self):
        """
        Return the dimension of the new cuspidal subspace, computed using
        dimension formulas.

        EXAMPLES::

            sage: m = ModularForms(GammaH(11,[2]), 2); m._dim_new_cuspidal()
            1
            sage: m = ModularForms(DirichletGroup(33).0,7); m
            Modular Forms space of dimension 26, character [-1, 1] and weight 7 over Rational Field
            sage: m._dim_new_cuspidal()
            20
            sage: m._dim_cuspidal()
            22
        """
        if isinstance(self.group(), Gamma1_class) and self.character() is not None:
            return self.group().dimension_new_cusp_forms(self.weight(), self.character())
        else:
            return self.group().dimension_new_cusp_forms(self.weight())

    @cached_method
    def _dim_new_eisenstein(self):
        """
        Return the dimension of the new Eisenstein subspace, computed
        by enumerating all Eisenstein series of the appropriate level.

        EXAMPLES::

            sage: m = ModularForms(Gamma0(11), 4)
            sage: m._dim_new_eisenstein()
            0
            sage: m = ModularForms(Gamma0(11), 2)
            sage: m._dim_new_eisenstein()
            1
            sage: m = ModularForms(DirichletGroup(36).0,5); m
            Modular Forms space of dimension 28, character [-1, 1] and weight 5 over Rational Field
            sage: m._dim_new_eisenstein()
            2
            sage: m._dim_eisenstein()
            8
        """
        if isinstance(self.group(), Gamma0_class) and self.weight() == 2:
            if is_prime(self.level()):
                d = 1
            else:
                d = 0
        else:
            E = self.eisenstein_series()
            d = len([g for g in E if g.new_level() == self.level()])
        return d

    ####################################################################
    # Computations of all Eisenstein series in self
    ####################################################################

    @cached_method
    def eisenstein_params(self):
        """
        Return parameters that define all Eisenstein series in ``self``.

        OUTPUT: an immutable Sequence

        EXAMPLES::

            sage: m = ModularForms(Gamma0(22), 2)
            sage: v = m.eisenstein_params(); v
            [(Dirichlet character modulo 22 of conductor 1 mapping 13 |--> 1, Dirichlet character modulo 22 of conductor 1 mapping 13 |--> 1, 2), (Dirichlet character modulo 22 of conductor 1 mapping 13 |--> 1, Dirichlet character modulo 22 of conductor 1 mapping 13 |--> 1, 11), (Dirichlet character modulo 22 of conductor 1 mapping 13 |--> 1, Dirichlet character modulo 22 of conductor 1 mapping 13 |--> 1, 22)]
            sage: type(v)
            <class 'sage.structure.sequence.Sequence_generic'>
        """
        eps = self.character()
        if eps is None:
            if isinstance(self.group(), Gamma1_class):
                eps = self.level()
            else:
                raise NotImplementedError
        params = eis_series.compute_eisenstein_params(eps, self.weight())
        return Sequence(params, immutable=True)

    def eisenstein_series(self):
        """
        Return all Eisenstein series associated to this space.

        ::

            sage: ModularForms(27,2).eisenstein_series()
            [q^3 + O(q^6),
             q - 3*q^2 + 7*q^4 - 6*q^5 + O(q^6),
             1/12 + q + 3*q^2 + q^3 + 7*q^4 + 6*q^5 + O(q^6),
             1/3 + q + 3*q^2 + 4*q^3 + 7*q^4 + 6*q^5 + O(q^6),
             13/12 + q + 3*q^2 + 4*q^3 + 7*q^4 + 6*q^5 + O(q^6)]

        ::

            sage: ModularForms(Gamma1(5),3).eisenstein_series()
            [-1/5*zeta4 - 2/5 + q + (4*zeta4 + 1)*q^2 + (-9*zeta4 + 1)*q^3 + (4*zeta4 - 15)*q^4 + q^5 + O(q^6),
             q + (zeta4 + 4)*q^2 + (-zeta4 + 9)*q^3 + (4*zeta4 + 15)*q^4 + 25*q^5 + O(q^6),
             1/5*zeta4 - 2/5 + q + (-4*zeta4 + 1)*q^2 + (9*zeta4 + 1)*q^3 + (-4*zeta4 - 15)*q^4 + q^5 + O(q^6),
             q + (-zeta4 + 4)*q^2 + (zeta4 + 9)*q^3 + (-4*zeta4 + 15)*q^4 + 25*q^5 + O(q^6)]

        ::

            sage: eps = DirichletGroup(13).0^2
            sage: ModularForms(eps,2).eisenstein_series()
            [-7/13*zeta6 - 11/13 + q + (2*zeta6 + 1)*q^2 + (-3*zeta6 + 1)*q^3 + (6*zeta6 - 3)*q^4 - 4*q^5 + O(q^6),
             q + (zeta6 + 2)*q^2 + (-zeta6 + 3)*q^3 + (3*zeta6 + 3)*q^4 + 4*q^5 + O(q^6)]
        """
        return self.eisenstein_submodule().eisenstein_series()

    def _compute_q_expansion_basis(self, prec):
        """
        EXAMPLES::

            sage: m = ModularForms(11,4)
            sage: m._compute_q_expansion_basis(5)
            [q + 3*q^3 - 6*q^4 + O(q^5), q^2 - 4*q^3 + 2*q^4 + O(q^5), 1 + O(q^5), q + 9*q^2 + 28*q^3 + 73*q^4 + O(q^5)]
        """
        S = self.cuspidal_submodule()
        E = self.eisenstein_submodule()
        B_S = S._compute_q_expansion_basis(prec)
        B_E = E._compute_q_expansion_basis(prec)
        return B_S + B_E

    def _compute_hecke_matrix(self, n):
        """
        Compute the matrix of the Hecke operator `T_n` acting on ``self``.

        .. NOTE::

            If ``self`` is a level 1 space, the much faster Victor Miller basis
            is used for this computation.

        EXAMPLES::

            sage: M = ModularForms(11, 2)
            sage: M._compute_hecke_matrix(6)
            [ 2  0]
            [ 0 12]

        Check that :issue:`22780` is fixed::

            sage: M = ModularForms(1, 12)
            sage: M._compute_hecke_matrix(2)
            [ -24    0]
            [   0 2049]
            sage: ModularForms(1, 2).hecke_matrix(2)
            []

        TESTS:

        The following Hecke matrix is 43x43 with very large integer entries.
        We test it indirectly by computing the product and the sum of its
        eigenvalues, and reducing these two integers modulo all the primes
        less than 100::

            sage: M = ModularForms(1, 512)
            sage: t = M._compute_hecke_matrix(5)     # long time (2s)
            sage: t[-1, -1] == 1 + 5^511             # long time (0s, depends on above)
            True
            sage: f = t.charpoly()                   # long time (4s)
            sage: [f[0]%p for p in prime_range(100)] # long time (0s, depends on above)
            [0, 0, 0, 0, 1, 9, 2, 7, 0, 0, 0, 0, 1, 12, 9, 16, 37, 0, 21, 11, 70, 22, 0, 58, 76]
            sage: [f[42]%p for p in prime_range(100)] # long time (0s, depends on above)
            [0, 0, 4, 0, 10, 4, 4, 8, 12, 1, 23, 13, 10, 27, 20, 13, 16, 59, 53, 41, 11, 13, 12, 6, 82]
        """
        if self.level() == 1:
            k = self.weight()
            d = self.dimension()
            if d == 0:
                return matrix(self.base_ring(), 0, 0, [])
            from sage.modular.all import victor_miller_basis, hecke_operator_on_basis
            vmb = victor_miller_basis(k, prec=d * n + 1)[1:]
            Tcusp = hecke_operator_on_basis(vmb, n, k)
            return Tcusp.block_sum(matrix(self.base_ring(), 1, 1,
                                          [sigma(n, k - 1)]))
        else:
            return space.ModularFormsSpace._compute_hecke_matrix(self, n)

    def _compute_hecke_matrix_prime_power(self, p, r):
        r"""
        Compute the Hecke matrix `T_{p^r}`, where `p` is prime and `r \ge 2`.

        This is an internal method.  End users are encouraged to use the
        method hecke_matrix() instead.

        TESTS:

            sage: M = ModularForms(1, 12)
            sage: M._compute_hecke_matrix_prime_power(5, 3)
            [           -359001100500                        0]
            [                       0 116415324211120654296876]
            sage: delta_qexp(126)[125]
            -359001100500
            sage: eisenstein_series_qexp(12, 126)[125]
            116415324211120654296876
        """
        if self.level() == 1:
            return self._compute_hecke_matrix(p**r)
        else:
            return space.ModularFormsSpace._compute_hecke_matrix_prime_power(self, p, r)

    def hecke_polynomial(self, n, var='x'):
        r"""
        Compute the characteristic polynomial of the Hecke operator `T_n` acting
        on this space. Except in level 1, this is computed via modular symbols,
        and in particular is faster to compute than the matrix itself.

        EXAMPLES::

            sage: ModularForms(17,4).hecke_polynomial(2)
            x^6 - 16*x^5 + 18*x^4 + 608*x^3 - 1371*x^2 - 4968*x + 7776

        Check that this gives the same answer as computing the actual Hecke
        matrix (which is generally slower)::

            sage: ModularForms(17,4).hecke_matrix(2).charpoly()
            x^6 - 16*x^5 + 18*x^4 + 608*x^3 - 1371*x^2 - 4968*x + 7776
        """
        return self.cuspidal_submodule().hecke_polynomial(n, var) * self.eisenstein_submodule().hecke_polynomial(n, var)



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.rings.finite_rings
r"""
Morphisms between Ore modules

Let `R` be a commutative ring, `\theta : R \to R` by a ring
endomorphism and `\partial : R \to R` be a `\theta`-derivation.
Let also `\mathcal S = R[X; \theta, \partial]` denote the
associated Ore polynomial ring.

By definition, a Ore module is a module over `\mathcal S`. In
SageMath, there are rather represented as modules over `R`
equipped with the map giving the action of the Ore variable `X`.
We refer to :mod:`sage.modules.ore_module` for more details.

A morphism of Ore modules is a `R`-linear morphism commuting
with the Ore action, or equivalenty a `\mathcal S`-linear map.

.. RUBRIC:: Construction of morphisms

There are several ways for creating Ore modules morphisms in SageMath.
First of all, one can use the method :meth:`sage.modules.ore_module.OreModule.hom`,
passing to it the matrix (in the canonical bases) of the morphism
we want to build::

    sage: K.<z> = GF(5^3)
    sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
    sage: M.<e0,e1> = S.quotient_module(X^2 + X + z)

    sage: mat = matrix(2, 2, [z,     3*z^2 + z + 2,
    ....:                     z + 1,       4*z + 4])
    sage: f = M.hom(mat)
    sage: f
    Ore module endomorphism of Ore module <e0, e1> over Finite Field in z of size 5^3 twisted by z |--> z^5

Clearly, this method is not optimal: typing all the entries of the
defining matrix is long and a potential source of errors.

Instead, one can use a dictionary encoding the values taken by the
morphism on a set of generators; the morphism is then automatically
prolonged by `\mathcal S`-linearity.
Actually here, `f` was just the multiplication by `X^3` on `M`.
We can then redefine it simply as follows::

    sage: g = M.hom({e0: X^3*e0})
    sage: g
    Ore module endomorphism of Ore module <e0, e1> over Finite Field in z of size 5^3 twisted by z |--> z^5

One can then recover the matrix by using the method
:meth:`sage.modules.ore_module_morphism.OreModuleMorphism.matrix`::

    sage: g.matrix()
    [            z 3*z^2 + z + 2]
    [        z + 1       4*z + 4]

Alternatively, one can use the method
:meth:`sage.modules.ore_module.OreModule.multiplication_map`::

    sage: h = M.multiplication_map(X^3)
    sage: h
    Ore module endomorphism of Ore module <e0, e1> over Finite Field in z of size 5^3 twisted by z |--> z^5
    sage: g == h
    True

Of course, this method also accepts values in the base ring::

    sage: h = M.multiplication_map(2)
    sage: h
    Ore module endomorphism of Ore module <e0, e1> over Finite Field in z of size 5^3 twisted by z |--> z^5
    sage: h.matrix()
    [2 0]
    [0 2]

Be careful that scalar multiplications do not always properly
define a morphism of Ore modules::

    sage: M.multiplication_map(z)
    Traceback (most recent call last):
    ...
    ValueError: does not define a morphism of Ore modules

.. RUBRIC: Kernels, images and related things

SageMath provides methods to compute kernels, cokernels,
images and coimages. In order to illustrate this, we will
build the sequence

.. MATH::

    0 \to \mathcal S/ \mathcal S P
    \to \mathcal S/ \mathcal S PQ
    \to \mathcal S/ \mathcal S Q \to 0

and check that it is exact.
We first build the Ore modules::

    sage: P = X^2 + z*X + 1
    sage: Q = X^3 + z^2*X^2 + X + z
    sage: U = S.quotient_module(P, names='u')
    sage: U.inject_variables()
    Defining u0, u1
    sage: V = S.quotient_module(P*Q, names='v')
    sage: V.inject_variables()
    Defining v0, v1, v2, v3, v4
    sage: W = S.quotient_module(Q, names='w')
    sage: W.inject_variables()
    Defining w0, w1, w2

Next, we build the morphisms::

    sage: f = U.hom({u0: Q*v0})
    sage: g = V.hom({v0: w0})

We can now check that `f` is injective by computing its
kernel::

    sage: f.kernel()
    Ore module of rank 0 over Finite Field in z of size 5^3 twisted by z |--> z^5

We see on the output that it has dimension `0`; so it
vanishes. Instead of reading the output, one can check
programmatically the vanishing of a Ore module using the
method :meth:`sage.modules.ore_module.OreModule.is_zero`::

    sage: f.kernel().is_zero()
    True

Actually, in our use case, one can, more simply, use the method
:meth:`sage.modules.ore_module_morphism.OreModuleMorphism.is_injective`::

    sage: f.is_injective()
    True

Similarly, one checks that `g` is surjective::

    sage: g.is_surjective()
    True

or equivalently::

    sage: g.cokernel().is_zero()
    True

Now, we need to check that the kernel of `g` equals the
image of `f`. For this, we compute both and compare the
results::

    sage: ker = g.kernel()
    sage: im = f.image()
    sage: ker == im
    True

As a sanity check, one can also verity that the composite
`g \circ f` vanishes::

    sage: h = g * f
    sage: h
    Ore module morphism:
      From: Ore module <u0, u1> over Finite Field in z of size 5^3 twisted by z |--> z^5
      To:   Ore module <w0, w1, w2> over Finite Field in z of size 5^3 twisted by z |--> z^5
    sage: h.is_zero()
    True

Let us now consider another morphism `f` and build the
canonical isomorphism `\text{coim }f \to \text{im }f`
that it induces.
We start by defining `f`::

    sage: A = X + z
    sage: B = X + z + 1
    sage: P = X^2 + X + z
    sage: U = S.quotient_module(B*P, names='u')
    sage: U.inject_variables()
    Defining u0, u1, u2
    sage: V = S.quotient_module(P*A, names='v')
    sage: V.inject_variables()
    Defining v0, v1, v2

    sage: f = U.hom({u0: A*v0})
    sage: f
    Ore module morphism:
      From: Ore module <u0, u1, u2> over Finite Field in z of size 5^3 twisted by z |--> z^5
      To:   Ore module <v0, v1, v2> over Finite Field in z of size 5^3 twisted by z |--> z^5

Now we compute the image and the coimage::

    sage: I = f.image(names='im')
    sage: I
    Ore module <im0, im1> over Finite Field in z of size 5^3 twisted by z |--> z^5

    sage: C = f.coimage(names='co')
    sage: C
    Ore module <co0, co1> over Finite Field in z of size 5^3 twisted by z |--> z^5

We can already check that the image and the coimage have the
same rank. We now want to construct the isomorphism between
them. For this, we first need to corestrict `f` to its image.
This is achieved via the method
:meth:`sage.modules.ore_module.OreSubmodule.morphism_corestriction`
of the Ore module::

    sage: g = I.morphism_corestriction(f)
    sage: g
    Ore module morphism:
      From: Ore module <u0, u1, u2> over Finite Field in z of size 5^3 twisted by z |--> z^5
      To:   Ore module <im0, im1> over Finite Field in z of size 5^3 twisted by z |--> z^5

Next, we want to factor `g` by the coimage. We proceed as follows::

    sage: h = C.morphism_quotient(g)
    sage: h
    Ore module morphism:
      From: Ore module <co0, co1> over Finite Field in z of size 5^3 twisted by z |--> z^5
      To:   Ore module <im0, im1> over Finite Field in z of size 5^3 twisted by z |--> z^5

We have found the morphism we were looking for: it is `h`.
We can now check that it is an isomorphism::

    sage: h.is_isomorphism()
    True

As a shortcut, we can use explicit conversions as follows::

    sage: H = Hom(C, I)  # the hom space
    sage: h2 = H(f)
    sage: h2
    Ore module morphism:
      From: Ore module <co0, co1> over Finite Field in z of size 5^3 twisted by z |--> z^5
      To:   Ore module <im0, im1> over Finite Field in z of size 5^3 twisted by z |--> z^5
    sage: h == h2
    True

.. RUBRIC: Determinants and characteristic polynomials

For endomorphisms, one can compute classical invariants as
determinants and characteristic polynomials.
To illustrate this, we check on an example that the characteristic
polynomial of the multiplication by `X^3` on the quotient
`\mathcal S / \mathcal S P` is the reduced norm of `P`::

    sage: P = X^5 + z*X^4 + z^2*X^2 + z + 1
    sage: M = S.quotient_module(P)
    sage: f = M.multiplication_map(X^3)
    sage: f.charpoly()
    x^5 + x^4 + x^3 + x^2 + 1

    sage: P.reduced_norm('x')
    x^5 + x^4 + x^3 + x^2 + 1

AUTHOR:

- Xavier Caruso (2024-10)
"""

# ***************************************************************************
#    Copyright (C) 2024 Xavier Caruso <xavier.caruso@normalesup.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 2 of the License, or
#    (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.misc.latex import latex
from sage.structure.element import Element
from sage.structure.factorization import Factorization
from sage.matrix.matrix0 import Matrix
from sage.matrix.constructor import matrix
from sage.categories.map import Map
from sage.categories.morphism import Morphism


class OreModuleMorphism(Morphism):
    r"""
    Generic class for morphism between Ore modules.
    """
    def __init__(self, parent, im_gens, check=True):
        r"""
        Initialize this Ore module.

        INPUT:

        - ``parent`` -- the hom space

        - ``im_gens`` -- the image of the generators (formatted as
          a list, a tuple, a dictionary or a matrix) or a Ore modules
          morphism

        - ``check`` (default: ``True``) -- a boolean, whether we
          should check if the given data correctly defined a morphism
          of Ore modules

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: type(f)
            <class 'sage.modules.ore_module_homspace.OreModule_homspace_with_category.element_class'>

            sage: TestSuite(f).run()

        An example with a nontrivial denominator::

            sage: from sage.modules.ore_module import OreModule
            sage: A.<t> = QQ[]
            sage: d = A.derivation()
            sage: M.<u> = OreModule(matrix([t^2 + 1]), d, denominator=t)
            sage: N.<v> = OreModule(matrix([t]), d)
            sage: M.hom({u: t*v})
            Ore module morphism:
              From: Ore module <u> over Univariate Polynomial Ring in t over Rational Field twisted by d/dt
              To:   Ore module <v> over Univariate Polynomial Ring in t over Rational Field twisted by d/dt
            sage: N.hom({v: t*u})
            Traceback (most recent call last):
            ...
            ValueError: does not define a morphism of Ore modules
        """
        Morphism.__init__(self, parent)
        domain = parent.domain()
        codomain = parent.codomain()
        base = domain.base_ring()
        fd = domain._pseudohom
        fc = codomain._pseudohom
        MS = parent.matrix_space()

        # Handle denominators
        dd = domain._denominator
        if dd is None:
            dd = Factorization([])
        dc = codomain._denominator
        if dc is None:
            dc = Factorization([])
        den = dd.lcm(dc)
        sd = den / dd
        sc = den / dc

        if (isinstance(im_gens, Element)
            and base.has_coerce_map_from(im_gens.parent())):
            self._matrix = MS(im_gens)
        elif isinstance(im_gens, Matrix):
            self._matrix = MS(im_gens)
        elif isinstance(im_gens, OreModuleMorphism):
            # Not optimal: too many intermediate morphisms constructed
            f = im_gens
            if f.domain() is not domain:
                f = domain._hom_change_domain(f)
            if f.codomain() is not codomain:
                f = codomain._hom_change_codomain(f)
            self._matrix = f._matrix
        elif isinstance(im_gens, (list, tuple)):
            if len(im_gens) != domain.rank():
                raise ValueError("wrong number of generators")
            self._matrix = matrix([codomain(v).list() for v in im_gens])
        elif isinstance(im_gens, dict):
            zero = parent.base_ring().zero()
            dimd = domain.rank()
            dimc = codomain.rank()
            d = dimc + dimd
            vs = [domain(x).list() + codomain(y).list() for x, y in im_gens.items()]
            if len(vs) < 2*d:
                vs += (2*d - len(vs)) * [d * [zero]]
            M = matrix(vs)
            M.echelonize()
            oldr = 0
            r = M.rank()
            iter = 1
            sd = base(sd.value())
            sc = base(sc.value())
            while r > oldr:
                for i in range(r):
                    row = M.row(i).list()
                    x = row[:dimd]
                    y = row[dimd:]
                    for _ in range(iter):
                        x = sd * fd(x)
                        y = sc * fc(y)
                    v = x.list() + y.list()
                    for j in range(d):
                        M[i+r,j] = v[j]
                M.echelonize()
                oldr = r
                r = M.rank()
                iter *= 2
            if list(M.pivots()) != list(range(dimd)):
                raise ValueError("does not define a morphism of Ore modules")
            self._matrix = M.submatrix(0, dimd, dimd, dimc)
        else:
            raise ValueError("cannot construct a morphism from the given data")
        if check:
            if isinstance(sd, Factorization):
                sd = base(sd.value())
                sc = base(sc.value())
            for x in parent.domain().basis():
                y = self._call_(x)
                fx = sd * fd(x)
                fy = sc * fc(y)
                if self._call_(fx) != fy:
                    raise ValueError("does not define a morphism of Ore modules")

    def _repr_type(self):
        r"""
        Return a string with the type of this morphism.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: f._repr_type()
            'Ore module'
        """
        return "Ore module"

    def _latex_(self):
        r"""
        Return a LaTeX representation of this morphism.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: latex(f)
            \begin{array}{l}
            \text{\texttt{Ore module morphism:}} \\
            \text{\texttt{{ }{ }From:}}\hspace{1ex} \texttt{Ore module of rank } 2\texttt{ over } \Bold{F}_{5^{3}} \texttt{ twisted by } z \mapsto z^{5} \\
            \text{\texttt{{ }{ }To:}}\hspace{3ex} \texttt{Ore module of rank } 2\texttt{ over } \Bold{F}_{5^{3}} \texttt{ twisted by } z \mapsto z^{5}
            \end{array}

        ::

            sage: Me = M.rename_basis('e')
            sage: fe = Me.multiplication_map(X^3)
            sage: latex(fe)
            \begin{array}{l}
            \text{\texttt{Ore module morphism:}} \\
            \text{\texttt{{ }{ }From:}}\hspace{1ex} \left<e_{0}, e_{1}\right>_{\Bold{F}_{5^{3}} , z \mapsto z^{5} } \\
            \text{\texttt{{ }{ }To:}}\hspace{3ex} \left<e_{0}, e_{1}\right>_{\Bold{F}_{5^{3}} , z \mapsto z^{5} }
            \end{array}
        """
        s = "\\begin{array}{l}\n"
        s += "\\text{\\texttt{%s morphism:}} \\\\\n" % self._repr_type()
        s += "\\text{\\texttt{{ }{ }From:}}\\hspace{1ex} %s \\\\\n" % latex(self.domain())
        s += "\\text{\\texttt{{ }{ }To:}}\\hspace{3ex} %s\n" % latex(self.codomain())
        s += "\\end{array}"
        return s

    def matrix(self):
        r"""
        Return the matrix defining this morphism.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(2)
            sage: f.matrix()
            [2 0]
            [0 2]

            sage: g = M.multiplication_map(X^3)
            sage: g.matrix()
            [            0 3*z^2 + z + 1]
            [      2*z + 1             0]
        """
        return self._matrix.__copy__()

    def _call_(self, x):
        r"""
        Return the image of `x` by this morphism.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M.<v,w> = S.quotient_module(X^2 + z*X + 1)
            sage: f = M.multiplication_map(X^3)
            sage: f(v)
            (2*z^2+4*z+4)*v + (4*z^2+3*z+3)*w

        We check that it is the correct answer::

            sage: X^3 * v
            (2*z^2+4*z+4)*v + (4*z^2+3*z+3)*w
        """
        return self.codomain()(x * self._matrix)

    def is_zero(self) -> bool:
        r"""
        Return ``True`` if this morphism is zero.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + z^2 + 1
            sage: M = S.quotient_module(P^2, names='e')
            sage: M.inject_variables()
            Defining e0, e1, e2, e3, e4, e5

            sage: f = M.hom({e0: P*e0})
            sage: f.is_zero()
            False
            sage: (f*f).is_zero()
            True
        """
        return self._matrix.is_zero()

    def is_identity(self) -> bool:
        r"""
        Return ``True`` if this morphism is the identity.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M.<v,w> = S.quotient_module(X^2 + z)
            sage: f = M.hom({v: v})
            sage: f.is_identity()
            True

            sage: f = M.hom({v: 2*v})
            sage: f.is_identity()
            False
        """
        return self.domain() is self.codomain() and self._matrix.is_one()

    def _add_(self, other):
        r"""
        Return the sum of this morphism and ``other``.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: g = M.multiplication_map(X^6)
            sage: h = f + g
            sage: h == M.multiplication_map(X^3 + X^6)
            True
        """
        if not isinstance(other, OreModuleMorphism):
            raise ValueError("the morphism is not a morphism of Ore modules")
        H = self.parent()
        return H(self._matrix + other._matrix, check=False)

    def _neg_(self):
        r"""
        Return the oppositive of this morphism.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: g = -f
            sage: g == M.multiplication_map(-X^3)
            True
        """
        H = self.parent()
        return H(-self._matrix, check=False)

    def _sub_(self, other):
        r"""
        Return the different between this morphism and ``other``.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: g = M.multiplication_map(X^6)
            sage: h = f - g
            sage: h == M.multiplication_map(X^3 - X^6)
            True
        """
        if not isinstance(other, OreModuleMorphism):
            raise ValueError("the morphism is not a morphism of Ore modules")
        H = self.parent()
        return H(self._matrix - other._matrix, check=False)

    def _rmul_(self, a):
        r"""
        Return the product of the scalar `a` by this morphism.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: g = 2*f
            sage: g == M.multiplication_map(2*X^3)
            True
        """
        H = self.parent()
        return H(a*self._matrix, check=False)

    def __eq__(self, other):
        r"""
        Return ``True`` if this morphism is equal to ``other``.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = 2*M.multiplication_map(X^3)
            sage: g = M.multiplication_map(2*X^3)
            sage: f == g
            True
        """
        if not isinstance(other, OreModuleMorphism):
            try:
                other = self.parent()(other)
            except ValueError:
                return False
        return self._matrix == other._matrix

    def is_injective(self) -> bool:
        r"""
        Return ``True`` if this morphism is injective.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + z^2
            sage: M = S.quotient_module(P^2, names='m')
            sage: M.inject_variables()
            Defining m0, m1, m2, m3, m4, m5
            sage: N = S.quotient_module(P, names='n')
            sage: N.inject_variables()
            Defining n0, n1, n2

            sage: f = N.hom({n0: P*m0})
            sage: f.is_injective()
            True

            sage: g = M.hom({m0: n0})
            sage: g.is_injective()
            False
        """
        return self.kernel().rank() == 0

    def is_surjective(self) -> bool:
        r"""
        Return ``True`` if this morphism is surjective.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + z^2
            sage: M = S.quotient_module(P^2, names='m')
            sage: M.inject_variables()
            Defining m0, m1, m2, m3, m4, m5
            sage: N = S.quotient_module(P, names='n')
            sage: N.inject_variables()
            Defining n0, n1, n2

            sage: f = N.hom({n0: P*m0})
            sage: f.is_surjective()
            False

            sage: g = M.hom({m0: n0})
            sage: g.is_surjective()
            True
        """
        return self.image()._submodule.basis.is_one()

    def is_bijective(self) -> bool:
        r"""
        Return ``True`` if this morphism is bijective.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: f.is_bijective()
            True

            sage: N = S.quotient_module(X^2)
            sage: g = N.multiplication_map(X^3)
            sage: g.is_bijective()
            False
        """
        return self.is_injective() and self.is_surjective()

    def is_isomorphism(self) -> bool:
        r"""
        Return ``True`` if this morphism is an isomorphism.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: f.is_isomorphism()
            True

            sage: N = S.quotient_module(X^2)
            sage: g = N.multiplication_map(X^3)
            sage: g.is_isomorphism()
            False
        """
        return self.is_bijective()

    def _composition_(self, other, homset):
        r"""
        Return the composite ``other`` `\circ` ``self``.

        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: g = f * f
            sage: g == M.multiplication_map(X^6)
            True
        """
        if not isinstance(other, OreModuleMorphism):
            raise ValueError("the morphism is not a morphism of Ore modules")
        return homset(other._matrix * self._matrix, check=False)

    def inverse(self):
        r"""
        Return the inverse of this morphism.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^2 + z
            sage: M.<e0,e1,e2,e3> = S.quotient_module(P^2)

            sage: f = M.multiplication_map(X^3)
            sage: g = f.inverse()
            sage: (f*g).is_identity()
            True
            sage: (g*f).is_identity()
            True

        If the morphism is not invertible, an error is raised::

            sage: h = M.hom({e0: P*e0})
            sage: h.inverse()
            Traceback (most recent call last):
            ...
            ValueError: this morphism is not invertible
        """
        if not self.is_isomorphism():
            raise ValueError("this morphism is not invertible")
        H = self.parent()
        return H(self._matrix.inverse(), check=False)

    __invert__ = inverse

    def kernel(self, names=None):
        r"""
        Return the kernel of this morphism.

        INPUT:

        - ``names`` (default: ``None``) -- the name of the vectors in a
          basis of the kernel

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + z^2
            sage: M = S.quotient_module(P^2, names='m')
            sage: M.inject_variables()
            Defining m0, m1, m2, m3, m4, m5
            sage: N = S.quotient_module(P, names='n')
            sage: N.inject_variables()
            Defining n0, n1, n2

            sage: f = M.hom({m0: n0})
            sage: ker = f.kernel()
            sage: ker
            Ore module of rank 3 over Finite Field in z of size 5^3 twisted by z |--> z^5
            sage: ker.basis()
            [m0 + (2*z^2+3*z+1)*m3 + (4*z^2+3*z+3)*m4 + (2*z^2+3*z)*m5,
             m1 + (z+3)*m3 + (z^2+z+4)*m4,
             m2 + (2*z^2+4*z+2)*m4 + (2*z^2+z+1)*m5]
        """
        mat = self._matrix
        if hasattr(mat, 'minimal_kernel_matrix'):
            # Optimisation
            ker = mat.minimal_kernel_matrix()
        else:
            ker = mat.left_kernel_matrix()
        domain = self.domain()
        return domain._submodule_class(domain, ker, False, names)

    def image(self, saturate=False, names=None):
        r"""
        Return the image of this morphism.

        INPUT:

        - ``saturate`` (default: ``False``) -- a boolean; if ``True``,
          return the saturation (in the codomain) of the image (see
          :meth:`sage.modules.ore_module.OreSubmodule.saturate` for
          more details)

        - ``names`` (default: ``None``) -- the name of the vectors in a
          basis of the image

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + z^2
            sage: M = S.quotient_module(P^2, names='m')
            sage: M.inject_variables()
            Defining m0, m1, m2, m3, m4, m5
            sage: N = S.quotient_module(P, names='n')
            sage: N.inject_variables()
            Defining n0, n1, n2

            sage: f = N.hom({n0: P*m0})
            sage: im = f.image()
            sage: im
            Ore module of rank 3 over Finite Field in z of size 5^3 twisted by z |--> z^5
            sage: im.basis()
            [m0 + (2*z^2+3*z+1)*m3 + (4*z^2+3*z+3)*m4 + (2*z^2+3*z)*m5,
             m1 + (z+3)*m3 + (z^2+z+4)*m4,
             m2 + (2*z^2+4*z+2)*m4 + (2*z^2+z+1)*m5]
        """
        codomain = self.codomain()
        return codomain._submodule_class(codomain, self._matrix, saturate, names)

    def cokernel(self, remove_torsion=False, names=None):
        r"""
        Return the cokernel of this morphism.

        INPUT:

        - ``remove_torsion`` (default: ``False``) -- a boolean;
          if ``True``, return the cokernel quotiented out by its
          torsion part

        - ``names`` (default: ``None``) -- the name of the vectors in a
          basis of the cokernel

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + z^2
            sage: M = S.quotient_module(P^2, names='m')
            sage: M.inject_variables()
            Defining m0, m1, m2, m3, m4, m5
            sage: N = S.quotient_module(P, names='n')
            sage: N.inject_variables()
            Defining n0, n1, n2

            sage: f = N.hom({n0: P*m0})
            sage: coker = f.cokernel()
            sage: coker
            Ore module of rank 3 over Finite Field in z of size 5^3 twisted by z |--> z^5
            sage: coker.basis()
            [m3, m4, m5]
        """
        codomain = self.codomain()
        return codomain._quotientModule_class(codomain, self._matrix, remove_torsion, names)

    def coimage(self, names=None):
        r"""
        Return the coimage of this morphism.

        INPUT:

        - ``names`` (default: ``None``) -- the name of the vectors in a
          basis of the coimage

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + z^2
            sage: M = S.quotient_module(P^2, names='m')
            sage: M.inject_variables()
            Defining m0, m1, m2, m3, m4, m5
            sage: N = S.quotient_module(P, names='n')
            sage: N.inject_variables()
            Defining n0, n1, n2

            sage: f = M.hom({m0: n0})
            sage: coim = f.coimage()
            sage: coim
            Ore module of rank 3 over Finite Field in z of size 5^3 twisted by z |--> z^5
            sage: coim.basis()
            [m3, m4, m5]
        """
        mat = self._matrix
        if hasattr(mat, 'minimal_kernel_matrix'):
            # Optimisation
            ker = mat.minimal_kernel_matrix()
        else:
            ker = mat.left_kernel_matrix()
        domain = self.domain()
        return domain._quotientModule_class(domain, ker, False, names)

    def determinant(self):
        r"""
        Return the determinant of this endomorphism.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: M.<m0,m1> = S.quotient_module(X^2 + z)
            sage: f = M.multiplication_map(X^3)
            sage: f.determinant()
            2

        If the domain differs from the codomain (even if they have
        the same rank), an error is raised::

            sage: N.<n0,n1> = S.quotient_module(X^2 + z^25)
            sage: g = M.hom({z*m0: n0})
            sage: g.determinant()
            Traceback (most recent call last):
            ...
            ValueError: determinants are only defined for endomorphisms
        """
        if self.domain() is not self.codomain():
            raise ValueError("determinants are only defined for endomorphisms")
        return self._matrix.determinant()

    det = determinant

    def characteristic_polynomial(self, var='x'):
        r"""
        Return the determinant of this endomorphism.

        EXAMPLES::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^3 + z*X^2 + (z^2 + 3)*X + z^5
            sage: M = S.quotient_module(P)
            sage: f = M.multiplication_map(X^3)
            sage: f.characteristic_polynomial()
            x^3 + x^2 + 2*x + 2

        We check that the latter is equal to the reduced norm
        of `P`::

            sage: P.reduced_norm('x')
            x^3 + x^2 + 2*x + 2

        TESTS::

            sage: M.<m0,m1> = S.quotient_module(X^2 + z)
            sage: N.<n0,n1> = S.quotient_module(X^2 + z^25)
            sage: g = M.hom({z*m0: n0})
            sage: g.characteristic_polynomial()
            Traceback (most recent call last):
            ...
            ValueError: characteristic polynomials are only defined for endomorphisms
        """
        if self.domain() is not self.codomain():
            raise ValueError("characteristic polynomials are only defined for endomorphisms")
        return self._matrix.charpoly(var)

    charpoly = characteristic_polynomial


class OreModuleRetraction(Map):
    r"""
    Conversion (partially defined) map from an ambient module
    to one of its submodule.
    """
    def _call_(self, y):
        r"""
        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^2 + z*X + 1
            sage: M = S.quotient_module(P^2, names='e')
            sage: M.inject_variables()
            Defining e0, e1, e2, e3
            sage: N = M.span(P*e0, names='u')
            sage: N(P*e0)  # indirect doctest
            u0 + z*u1

            sage: N(e0)
            Traceback (most recent call last):
            ...
            ValueError: not in the submodule
        """
        X = self.codomain()
        submodule = X._submodule
        rank = submodule.rank
        xs = y * submodule.coordinates
        if xs[rank:]:
            raise ValueError("not in the submodule")
        try:
            xs = xs[:rank].change_ring(X.base_ring())
        except (ValueError, TypeError):
            raise ValueError("not in the submodule")
        return X(xs)


class OreModuleSection(Map):
    r"""
    Section map of the projection onto a quotient.
    It is not necessarily compatible with the Ore action.
    """
    def _call_(self, y):
        r"""
        TESTS::

            sage: K.<z> = GF(5^3)
            sage: S.<X> = OrePolynomialRing(K, K.frobenius_endomorphism())
            sage: P = X^2 + z*X + 1
            sage: M = S.quotient_module(P^2, names='e')
            sage: M.inject_variables()
            Defining e0, e1, e2, e3
            sage: N = M.quo(P*e0, names='u')
            sage: N.inject_variables()
            Defining u0, u1
            sage: M(u0)  # indirect doctest
            e2
        """
        X = self.codomain()
        Y = self.domain()
        return X(y * Y._submodule.complement)



# sage_setup: distribution = sagemath-modules
# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_ntuple.h":
  ctypedef struct gsl_ntuple

  ctypedef struct gsl_ntuple_select_fn:
    int (* function) (void * ntuple_data, void * params)
    void * params

  ctypedef struct gsl_ntuple_value_fn:
    double (* function) (void * ntuple_data, void * params)
    void * params

  gsl_ntuple * gsl_ntuple_open (char * filename, void * ntuple_data, size_t size)

  gsl_ntuple * gsl_ntuple_create (char * filename, void * ntuple_data, size_t size)

  int gsl_ntuple_write (gsl_ntuple * ntuple)
  int gsl_ntuple_read (gsl_ntuple * ntuple)

  int gsl_ntuple_bookdata (gsl_ntuple * ntuple)

  int gsl_ntuple_project (gsl_histogram * h, gsl_ntuple * ntuple,
                          gsl_ntuple_value_fn *value_func,
                          gsl_ntuple_select_fn *select_func)

  int gsl_ntuple_close (gsl_ntuple * ntuple)

